#ifndef ccbcommon_h
#define ccbcommon_h

#ifndef __KERNEL__
#include <time.h>    /* time_t */
#endif

#include "ccbconstants.h"

/*-----------------------------------------------------------------------*
 * This is the public header file of the libccbcom library. It describes *
 * public functions, datatypes and constants that are common to both     *
 * the server and manager ends of the CCB communications library.        *
 *                                                                       *
 * In addition, the declarations within the first part of the file are   *
 * also used by the main CCB device driver.                              *
 *-----------------------------------------------------------------------*/

/* Enumerate the supported logging levels */

typedef enum {
  CCB_INFO,     /* A purely informational message */
  CCB_NOTICE,   /* A note about a probably inconsequential event */
  CCB_WARNING,  /* A warning about a potentially problematic event */
  CCB_ERROR,    /* A report of an event that requires operator attention */
  CCB_FAULT,    /* A report of a condition that is corrupting data */
  CCB_FATAL     /* A report of a system-wide failure */
} CCBLogLevel;

/*
 * Whenever a new ccb_log_server_msg() or ccb_log_client_msg()
 * statement is written, its id argument should be set to CCB_LOGID().
 * This allows scripts used by the CCB makefile to search for, and
 * subsequently fill in, the IDs of new log statements.  These scripts
 * simply place the newly assigned ID as the argument of the
 * CCB_LOGID() macro, which although it does nothing but echo its
 * argument, is retained, so that other scripts can search for it when
 * figuring out which IDs have been used so far.
 */
#define CCB_LOGID(x) (x)

/*
 * Enumerate the two phase switches, using the values of single
 * bits in a hypothetical unsigned integer to represent them.
 */
typedef enum {
  CCB_NO_PHASE_SWITCHES = 0,     /* Neither of the phase switches */
  CCB_PHASE_SWITCH_A = 1,        /* phase switch A */
  CCB_PHASE_SWITCH_B = 2,        /* phase switch B */
  CCB_ALL_PHASE_SWITCHES =       /* Both phase switches */
    CCB_PHASE_SWITCH_A | CCB_PHASE_SWITCH_B
} CCBPhaseSwitches;

/*
 * Enumerate the two calibration diodes, using the values of single
 * bits in a hypothetical unsigned integer to represent them.
 */
typedef enum {
  CCB_NO_CAL_DIODES = 0,      /* Neither of the calibration diodes */
  CCB_CAL_DIODE_A = 1,        /* Calibration diode A */
  CCB_CAL_DIODE_B = 2,        /* Calibration diode B */
  CCB_ALL_CAL_DIODES =        /* Both calibration diodes */
    CCB_CAL_DIODE_A | CCB_CAL_DIODE_B
} CCBCalDiodes;

/*
 * Enumerate the supported types of digitized samples.
 */
typedef enum {
  CCB_ADC_SAMPLES,   /* Raw ADC samples */
  CCB_FAKE_SAMPLES   /* Pseudo-random samples */
} CCBSampleType;

/*
 * Group the calibration-diode configuration parameters.
 */
typedef struct {
  unsigned short ncal;                      /* The number of cal steps */
  unsigned short diode_states[CCB_MAX_NCAL];/* The ncal cal-diode states, */
                                            /*  with each state being a */
                                            /*  bitwise union of CCBCalDiode */
                                            /*  enumerators. */
  unsigned long diode_times[CCB_MAX_NCAL];  /* The duration of each */
				            /*  cal-diode state, as the */
                                            /*  corresponding number of */
                                            /*  integrations. */
} CCBCalDiodeCnf;

/*
 * Group the phase-switch configuration parameters.
 */
typedef struct {
  unsigned short active_switches;  /* Which switches are active? This is */
                                   /*  a bitwise union of CCBPhaseSwitches */
                                   /*  enumerators. */
  unsigned short closed_switches;  /* Which switches start closed? This is */
                                   /*  a bitwise union of CCBPhaseSwitches */
                                   /*  enumerators. */
  unsigned short samp_per_state;   /* Samples per phase-switch state */
} CCBPhaseSwitchCnf;

/*
 * Group the timing configuration parameters.
 */
typedef struct {
  unsigned short phase_switch_dt;  /* The settling time of the phase */
				   /*  switches. */
  unsigned long diode_rise_dt;     /* The rise time of a cal diode */
  unsigned long diode_fall_dt;     /* The fall time of a cal diode */
  unsigned long integ_period;      /* The integration period */
  unsigned short roundtrip_dt;     /* The delay between toggling a */
                                   /*  receiver control line and */
                                   /*  the first effects reaching */
                                   /*  the CCB integrators. */
  unsigned short holdoff_dt;       /* The minimum time between */
                                   /*  generating CPU interrupts */
                                   /*  is 25.6us*(holdoff_dt+1). */
  unsigned short adc_delay_dt;     /* The offset of the ADC clock */
                                   /*  from the main FPGA clock, */
                                   /*  expressed as a multiple of */
                                   /*  10ns. */
} CCBTimingCnf;

/*
 * Group the sampler configuration parameters.
 */
typedef struct {
  unsigned short sample_type;   /* A CCBSampleType enumerator */
} CCBSamplerCnf;

/*-----------------------------------------------------------------------
 * The remainder of this file declares features that can not used by the
 * CCB device driver.
 *-----------------------------------------------------------------------*/

#ifndef __KERNEL__

/*
 * Times of events in the CCB are reported encapsulated in structures of
 * the following type.
 */
typedef struct {
  unsigned long mjd;    /* The Modified Julian Day number */
  unsigned long sec;    /* The number seconds into the day */
  unsigned long ns;     /* The number of nano-seconds after 'sec' */
} CCBTimeStamp;

/*
 * Provide a means to express long durations with nanosecond precision.
 */
typedef struct {       /* A time interval (t=sec+1.0e9*ns) */
  unsigned long sec;   /* The number of complete seconds */
  unsigned long ns;    /* The number of nano-seconds */
} CCBInterval;

/*
 * Lookup the current date and time, and record it in *t.
 */
int ccb_get_timestamp(CCBTimeStamp *t);

/*
 * Convert a date-and-time specified via a standard time_t variable,
 * plus a seconds fraction, to a CCB timestamp.
 */
int ccb_time_to_timestamp(time_t t, unsigned long ns, CCBTimeStamp *ts);

/*
 * Convert the time-of-day in a timestamp to hours, minutes and seconds.
 */
void ccb_hms_of_timestamp(CCBTimeStamp ts, unsigned *hours, unsigned *mins,
			 unsigned *secs);

/*
 * Compare two timestamps, returning -1, 0 or 1, depending on whether
 * ta < tb, ta==tb, or ta > tb respectively.
 */
int ccb_compare_timestamps(CCBTimeStamp ta, CCBTimeStamp tb);

/*
 * Return the amount of time remaining until a specified time, returning
 * zero if the time has already passed.
 */
int ccb_time_until(CCBTimeStamp ts, CCBInterval *dt);

/*
 * Add a time interval to a timestamp.
 */
int ccb_add_to_timestamp(const CCBTimeStamp *ta, const CCBInterval *dt,
			 CCBTimeStamp *tb);

/*
 * Enumerate the completion statuses that can be reported in control-command
 * acknowledgment replies.
 */
typedef enum {
  CCB_CMD_ACCEPTED,    /* This enumerator isn't forwarded to error callbacks */
  CCB_CMD_GARBLED,     /* The contents of the command message */
                       /*  were invalid and couldn't be fixed. */
  CCB_CMD_IGNORED,     /* The command didn't make sense at this */
                       /*  time, and was ignored. */
  CCB_CMD_SYSERR       /* An unexpected internal software or hardware */
		       /*  error was encountered while attempting */
		       /*  to execute the command. */
} CCBCmdStatus;

/*
 * Enumerate various conditions in the CCB server, with each condition
 * being represented by the value of a single bit.
 */
typedef enum {
  CCB_LINK_DOWN   = 1,  /* The telemetry link is down */
  CCB_BUFFER_FULL = 2,  /* The telemetry output buffer filled */
                        /*  up and hasn't drained yet, so data */
                        /*  are being discarded. */
  CCB_HARD_FAULT  = 4,  /* A hardware fault has been detected */
  CCB_SOFT_FAULT  = 8   /* A software fault has been detected */
} CCBGeneralStatus;

/*
 * Enumerate the types of data streams that can be sent over the
 * telemetry link by a CCB server. Each stream is represented by
 * the value of a single bit within a larger integer.
 */
typedef enum {
  CCB_INTEG_STREAM = 1,   /* The stream of integrated data */
  CCB_MONITOR_STREAM = 2, /* The stream of monitoring data */
  CCB_LOG_STREAM = 4,     /* The stream of log messages */
  CCB_NO_STREAMS = 0,     /* None of the above streams */
/* All of the above streams */
  CCB_ALL_STREAMS = CCB_INTEG_STREAM | CCB_MONITOR_STREAM | CCB_LOG_STREAM
} CCBTelemetryStream;

/*
 * CCB configuration is performed using opaque objects of the following
 * type.
 */
typedef struct CCBConfig CCBConfig;

/*
 * Create a new CCBConfig object, initialized either with the contents
 * of another CCBConfig object specified by the 'init' argument, or
 * if this is NULL, with the default power-on configuration of the CCB.
 */
CCBConfig *new_CCBConfig(void);

/*
 * Delete a CCBConfig object.
 */
CCBConfig *del_CCBConfig(CCBConfig *cnf);

/*
 * Reinitialize a CCBConfig object with the default power-on configuration
 * of the CCB.
 */
int ccb_default_config(CCBConfig *cnf);

/*
 * Make a copy of a CCB configuration object.
 */
int ccb_copy_config(const CCBConfig *orig, CCBConfig *dest);

/*
 * Check the validity of a CCBConfig configuration object.
 */
int ccb_check_config(CCBConfig *cnf, size_t errdim, char *errmsg);
/*
 * Report the set of configuration groups that differ between two CCB
 * configuration object. The return value is a bitmask union of
 * CCBConfigType enumerators.
 */
unsigned ccb_diff_config(const CCBConfig *ca, const CCBConfig *cb);

/*
 * Enumerate the types of configuration within a CCBConfig object.
 * Note that all enumerators have power-of-two values, such that
 * they can be used in a bitmask union.
 */
typedef enum {
  CCB_CNF_PHASE_SWITCHES = 1,  /* The phase-switch configuration parameters */
  CCB_CNF_CAL_DIODES     = 2,  /* The cal-diode configuration parameters */
  CCB_CNF_TIMING         = 4,  /* The timing configuration parameters */
  CCB_CNF_SAMPLER        = 8,  /* The sampler configuration parameters */
/*
 * The union of all of the above enumerators.
 */
  CCB_CNF_ALL = CCB_CNF_PHASE_SWITCHES | CCB_CNF_CAL_DIODES |
                CCB_CNF_TIMING | CCB_CNF_SAMPLER
} CCBConfigType;


/*
 * Change the phase-switch configuration in a given CCB configuration object.
 */
int ccb_set_phase_switch_cnf(CCBConfig *cnf,
			     unsigned short active_switches,
			     unsigned short closed_switches,
			     unsigned short samp_per_state);

/*
 * Change the cal-diode configuration in a given CCB configuration object.
 */
int ccb_set_cal_diode_cnf(CCBConfig *cnf,
			  unsigned short ncal,
			  const unsigned short *diode_states,
			  const unsigned long *diode_times);

/*
 * Change the timing configuration in a given CCB configuration object.
 */
int ccb_set_timing_cnf(CCBConfig *cnf,
		       unsigned short phase_switch_dt,
		       unsigned long diode_rise_dt,
		       unsigned long diode_fall_dt,
		       unsigned long integ_period,
		       unsigned short roundtrip_dt,
		       unsigned short holdoff_dt,
		       unsigned short adc_delay_dt);

/*
 * Change the sampling configuration in a given CCB configuration object.
 */
int ccb_set_sampler_cnf(CCBConfig *cnf, CCBSampleType sample_type);

/*
 * Query the phase-switch configuration in a given CCB configuration object.
 */
int ccb_get_phase_switch_cnf(const CCBConfig *cnf, CCBPhaseSwitchCnf *pars);

/*
 * Query the cal-diode configuration in a given CCB configuration object.
 */
int ccb_get_cal_diode_cnf(const CCBConfig *cnf, CCBCalDiodeCnf *pars);

/*
 * Query the timing configuration in a given CCB configuration object.
 */
int ccb_get_timing_cnf(const CCBConfig *cnf, CCBTimingCnf *pars);

/*
 * Query the sampler configuration in a given CCB configuration object.
 */
int ccb_get_sampler_cnf(const CCBConfig *cnf, CCBSamplerCnf *pars);

/*
 * Set a CCBConfig object to a given configuration.
 */
int ccb_set_config(CCBConfig *cnf,
		   const CCBPhaseSwitchCnf *phase,
		   const CCBCalDiodeCnf *cal,
		   const CCBTimingCnf *timing,
		   const CCBSamplerCnf *sampler);

/*
 * Scale a time interval by an unsigned long integer.
 */
int ccb_scale_interval(CCBInterval *dt, unsigned long factor,
		       CCBInterval *ans);

/*
 * Compute the sum of two time intervals.
 */
int ccb_add_intervals(CCBInterval *dt1, const CCBInterval *dt2,
		      CCBInterval *sum);

/*
 * Subtract time-interval dt2 from time-interval dt1.
 */
int ccb_subtract_interval(CCBInterval *dt1, CCBInterval *dt2,
			  CCBInterval *dif);

/*
 * Compare two time intervals, returning -1 if dt1<dt2, 0 if dt1==dt2,
 * or 1 if dt1>dt2.
 */
int ccb_compare_intervals(const CCBInterval *dt1, const CCBInterval *dt2);

/*
 * Initialize a time interval to zero.
 */
void ccb_zero_interval(CCBInterval *dt);

/*
 * Return non-zero if the specified interval is zero.
 */
int ccb_interval_is_zero(CCBInterval *dt);

/*
 * Convert from a number of clock ticks to the corresponding time interval.
 */
void ccb_clock_interval(unsigned long ticks, CCBInterval *dt);

/*
 * Return the calibration diode settling-time compensation delay for
 * a given integration-period within a fully-described scan.
 */
int ccb_cal_diode_delay(const CCBConfig *cnf, unsigned long integration,
			CCBInterval *dt);

/*
 * Return the dominating settling time of a given combination of
 * simultaneously commanded cal-diode and phase-switch transitions.
 */
int ccb_settling_time(const CCBConfig *cnf,
		      unsigned prev_cal, unsigned next_cal,
		      unsigned prev_phs, unsigned next_phs,
		      CCBInterval *dt);

/*
 * Return the duration of an integration.
 */
int ccb_integration_duration(const CCBConfig *cnf, CCBInterval *dt);

/*
 * Return the number of measurements per phase-switch cycle according
 * to the set of phase-switches that are active.
 */
unsigned ccb_cycle_length(unsigned short active_switches);

/*
 * Return the predicted duration of a fully-described scan of a given
 * number of integration periods.
 */
int ccb_scan_duration(const CCBConfig *cnf, unsigned long integrations,
		      CCBInterval *scan_dt);

/*
 * Return the number of integrations per cal-diode cycle.
 */
int ccb_cal_cycle_length(const CCBConfig *cnf, unsigned long *ninteg);

/*
 * Compute the effective integration-time of a single integration period.
 */
int ccb_integration_time(const CCBConfig *cnf, CCBInterval *dt);

/*
 * Compute the number of complete integrations, and remaining time,
 * that will fit within a given time duration.
 */
int ccb_integ_per_interval(const CCBConfig *cnf, CCBInterval *duration,
			   unsigned long *n, CCBInterval *rdt);

/*
 * Enumerate the two CCB links with the values of individual bits, so
 * that they can be placed within bitwise unions.
 */
typedef enum {
  CCB_CNTRL_LINK = 1,  /* The link to the CCB telemetry server */
  CCB_TELEM_LINK = 2,  /* The link to the CCB control server */
  CCB_ALL_LINKS = CCB_CNTRL_LINK | CCB_TELEM_LINK
} CCBLinkType;

/*
 * List the available CCB device drivers.
 */
typedef enum {
  CCB_NORMAL_DRIVER,  /* The real CCB device-driver */
  CCB_VIRTUAL_DRIVER  /* The CCB simulation device-driver */
} CCBDriverType;

/*
 * Enumerate the individual status bits that are included in the 'flags'
 * header parameter of integrations.
 */
typedef enum {
  CCB_CAL_A_ON = 1,   /* Included if cal-diode A was on */
  CCB_CAL_B_ON = 2,   /* Included if cal-diode B was on */
  CCB_INTEG_OK = 4,   /* Included if the integration is usable */
  CCB_SLAVE0_OK = 8,  /* Included if slave-FPGA 0 was present */
  CCB_SLAVE1_OK = 16, /* Included if slave-FPGA 1 was present */
  CCB_SLAVE2_OK = 32, /* Included if slave-FPGA 2 was present */
  CCB_SLAVE3_OK = 64  /* Included if slave-FPGA 3 was present */
} CCBIntegFlags;

#endif  /* !defined(__KERNEL__) */

#endif
