#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <stddef.h>

#include "ccbcontrolserver.h"
#include "netbuf.h"
#include "ccbnetobj.h"
#include "ccbconstants.h"
#include "ccbqueue.h"
#include "ccbcontrolmsg.h"
#include "ccbserverpriv.h"

/*
 * Define the object that manages the CCB server end of the control link.
 */
struct CCBControlServer {
  CCBControlServerStatus state; /* The operational status of the */
                                /*  control server. */
  CCBServerLink *server;            /* The parent CCB server */
  NetSendStr *nss;              /* The network output stream */
  NetReadStr *nrs;              /* The network input stream */
  CCBMessageQueue *rq;          /* The queue of outgoing messages */
  CCBConfig *cnf;               /* The requested CCB configuration */
};

static int cs_fill_control_reply_header(CCBControlReplyHeader *head,
					CCBControlReplyType type);
static int cs_deliver_command(CCBControlServer *cs, CCBControlCommand *cmd);
static int cs_stage_reply(CCBControlServer *cs, CCBControlReplyHeader *head);
static int cs_queue_control_reply(CCBControlServer *cs,
				  CCBControlReplyHeader *head);
static int cs_unqueue_next_reply(CCBControlServer *cs, CCBControlReply *msg);
static int cs_queue_cmd_ack(CCBControlServer *cs, unsigned long id,
			    CCBCmdStatus status);
static int cs_queue_cntrl_ping_reply(CCBControlServer *cs, unsigned long seq);
static int cs_queue_status_reply(CCBControlServer *cs,
				      unsigned long status);
static int ccb_cs_stage_scan(CCBControlServer *cs, unsigned long scan,
			     unsigned long mjd, unsigned long secs);

/*.......................................................................
 * Create a new CCBControlServer object.
 *
 * Input:
 *  server         CCBServerLink *  The parent CCB server.
 * Output:
 *  return  CCBControlServer *  The new object, or NULL on error.
 */
CCBControlServer *new_CCBControlServer(CCBServerLink *server)
{
  CCBControlServer *cs;  /* The object to be returned */
  char errmsg[200];      /* An error message */
/*
 * Check the arguments.
 */
  if(!server) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(1),
		"new_CCBControlServer: NULL argument(s).\n");
    return NULL;
  };
/*
 * Check that the tables of control message descriptions are valid.
 */
  if(ccb_check_net_table(&ccb_control_cmd_table, errmsg,
			 sizeof(errmsg))) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(2),
		       "Control command-table: %s\n", errmsg);
    return NULL;
  };
  if(ccb_check_net_table(&ccb_control_reply_table, errmsg,
			 sizeof(errmsg))) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(3),
		       "Control reply-table: %s\n", errmsg);
    return NULL;
  };
/*
 * Allocate the container.
 */
  cs = malloc(sizeof(CCBControlServer));
  if(!cs) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(4),
		"new_CCBControlServer: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBControlServer().
 */
  cs->state = CCB_CONTROL_SERVER_DOWN;
  cs->server = server;
  cs->nss = NULL;
  cs->nrs = NULL;
  cs->rq = NULL;
  cs->cnf = NULL;
/*
 * Get a CCB configuration object, initialized with the power-on
 * default configuration of the CCB.
 */
  cs->cnf = new_CCBConfig();
  if(!cs->cnf)
    return del_CCBControlServer(cs);
/*
 * Allocate the output stream manager.
 */
  cs->nss = new_NetSendStr(-1, ccb_net_max_obj_size(&ccb_control_reply_table));
  if(!cs->nss) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(5),
	"new_CCBControlServer: Insufficient memory for output stream.\n");
    return del_CCBControlServer(cs);
  };
/*
 * Allocate the input stream manager.
 */
  cs->nrs = new_NetReadStr(-1, ccb_net_max_obj_size(&ccb_control_cmd_table));
  if(!cs->nrs) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(6),
	"new_CCBControlServer: Insufficient memory for input stream.\n");
    return del_CCBControlServer(cs);
  };
/*
 * Allocate the message queue that is used to queue outgoing control reply
 * messages.
 */
  cs->rq = new_CCBMessageQueue(sizeof(CCBControlReply), CCB_CNTRL_REPLY_QLEN,
			       CCB_MQ_WRAP);
  if(!cs->rq) {
    ccb_log_server_msg(server, CCB_ERROR, CCB_LOGID(7),
	"new_CCBControlServer: Insufficient memory for message queue.\n");
    return del_CCBControlServer(cs);
  };
  return cs;
}

/*.......................................................................
 * Delete a CCBControlServer object.
 *
 * Input:
 *  cs     CCBControlServer *  The object to be deleted.
 * Output:
 *  return CCBControlServer *  The deleted object (always NULL).
 */
CCBControlServer *del_CCBControlServer(CCBControlServer *cs)
{
  if(cs) {
    cs->nss = del_NetSendStr(cs->nss);
    cs->nrs = del_NetReadStr(cs->nrs);
    cs->rq = del_CCBMessageQueue(cs->rq);
    cs->cnf = del_CCBConfig(cs->cnf);
    free(cs);
  };
  return NULL;
}

/*.......................................................................
 * Queue a control-link ping reply message to be sent to the current
 * manager.
 *
 * Input:
 *  cs    CCBControlServer *  The control-link server.
 *  seq      unsigned long    The sequence number of the ping command.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int cs_queue_cntrl_ping_reply(CCBControlServer *cs, unsigned long seq)
{
  CCBCntrlPingReply msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(8),
		"cs_queue_cntrl_ping_reply: NULL argument(s)\n");
    return 1;
  };
/*
 * Silently discard link-replies while the control link is down.
 */
  if(cs->state == CCB_CONTROL_SERVER_DOWN)
    return 0;
/*
 * Compose the reply.
 */
  if(cs_fill_control_reply_header(&msg.head, CCB_CNTRL_PING_REPLY))
    return 1;
/*
 * Fill in the body of the message.
 */
  msg.seq = seq;
/*
 * If no message is currently staged for transmission, stage the new message.
 * Otherwise append it to queue of outgoing control replies.
 */
  if(cs_queue_control_reply(cs, &msg.head))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a control-status reply message to be sent to the current manager.
 *
 * Input:
 *  cs    CCBControlServer *  The control-link server.
 *  status   unsigned long    A bitwise union of CCBGeneralStatus
 *                            enumerators.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int cs_queue_status_reply(CCBControlServer *cs,
				      unsigned long status)
{
  CCBStatusReply msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(9),
		"cs_queue_status_reply: NULL argument(s)\n");
    return 1;
  };
/*
 * Silently discard status-replies while the control link is down.
 */
  if(cs->state == CCB_CONTROL_SERVER_DOWN)
    return 0;
/*
 * Compose the reply.
 */
  if(cs_fill_control_reply_header(&msg.head, CCB_STATUS_REPLY))
    return 1;
  msg.status = status;
/*
 * If no message is currently staged for transmission, stage the new message.
 * Otherwise append it to the queue of outgoing control replies.
 */
  if(cs_queue_control_reply(cs, &msg.head))
    return 1;
  return 0;
}

/*.......................................................................
 * Queue a control-command acknowledgment message to be sent to the
 * current manager.
 *
 * Input:
 *  cs    CCBControlServer *  The control-link server.
 *  id       unsigned long    The manager-specified ID of the command
 *                            that is being acknowledged.
 *  status   CCBCmdStatus     The completion status of the command.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int cs_queue_cmd_ack(CCBControlServer *cs, unsigned long id,
			    CCBCmdStatus status)
{
  CCBCntrlCmdAck msg;   /* The new message */
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(10),
		"ccb_cs_queue_cmd_ack: NULL argument(s)\n");
    return 1;
  };
/*
 * Silently discard status-replies while the control link is down.
 */
  if(cs->state == CCB_CONTROL_SERVER_DOWN)
    return 0;
/*
 * Compose the reply.
 */
  if(cs_fill_control_reply_header(&msg.head, CCB_CNTRL_CMD_ACK))
    return 1;
  msg.id = id;
  msg.status = status;
/*
 * If no message is currently staged for transmission, stage the new message.
 * Otherwise append it to the queue of outgoing control replies.
 */
  if(cs_queue_control_reply(cs, &msg.head))
    return 1;
  return 0;
}

/*.......................................................................
 * Set up the contents of the header of a control reply message.
 *
 * Input:
 *  head  CCBControlReplyHeader *  The header to be initilized.
 *  type  CCBControlReplyType      The type of message.
 * Output:
 *  return               int    0 - OK.
 *                              1 - Error.
 */
static int cs_fill_control_reply_header(CCBControlReplyHeader *head,
					CCBControlReplyType type)
{
  head->base.type = type;
  return 0;
}

/*.......................................................................
 * Queue a message for transmission to the manager over the control link.
 * This involves either adding it to a holding queue, or encoding it for
 * transmission, depending on whether or not the server is currently
 * busy transmitting a previous message.
 *
 * Input:
 *  cs         CCBControlServer *  The control server.
 *  head  CCBControlReplyHeader *  The header of a new message to queue,
 *                                 or NULL to use the highest priority
 *                                 of any currently queued messages.
 * Output:
 *  return                  int    0 - OK.
 *                                 1 - Error.
 */
static int cs_queue_control_reply(CCBControlServer *cs,
				  CCBControlReplyHeader *head)
{
  CCBControlReply msg;  /* A dequeued message */
/*
 * Is the server ready to send a new message?
 */
  if(cs->state == CCB_CONTROL_SERVER_IDLE) {
/*
 * If no message was specified, get the highest priority message
 * from the queues.
 */
    if(!head) {
      if(cs_unqueue_next_reply(cs, &msg))
	return 0;       /* There were no messages waiting to be sent */
      head = &msg.head;
    };
/*
 * Stage the message for transmission.
 */
    if(cs_stage_reply(cs, head))
      return 1;
/*
 * If the server isn't ready to send a new message, but a new message
 * was specified, queue this message for later transmission.
 */
  } else if(head) {
/*
 * Lookup the size of the message structure that the header belongs to.
 */
    size_t size = ccb_net_native_size(&ccb_control_reply_table,
				      head->base.type);
    if(!size) {
      ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(11),
		  "cs_queue_control_reply: Unknown type of message.\n");
      return 1;
    };
/*
 * Append the message to the queue.
 */
    (void) ccb_mq_push(cs->rq, head, size);
  };
  return 0;
}

/*.......................................................................
 * Unqueue the next queued reply to be sent to the manager.
 *
 * Input:
 *  cs     CCBControlServer *  The control server.
 *  msg     CCBControlReply *  The reply message will be recorded in *msg.
 * Output:
 *  return                int    0 - OK.
 *                               1 - No message returned.
 */
static int cs_unqueue_next_reply(CCBControlServer *cs, CCBControlReply *msg)
{
  if(ccb_mq_pop(cs->rq, msg, sizeof(*msg)) == sizeof(*msg))
    return 0;
/*
 * There are no messages in the queue.
 */
  return 1;
}

/*.......................................................................
 * Stage a control reply for subsequent transmission to the manager.
 *
 * Input:
 *  cs          CCBControlServer *  The control server.
 *  head   CCBControlReplyHeader *  The header of a particular
 *                                  control message to stage for
 *                                  transmission.
 * Output:
 *  return                   int    0 - OK.
 *                                  1 - Error.
 */
static int cs_stage_reply(CCBControlServer *cs, CCBControlReplyHeader *head)
{
/*
 * Mark the server as idle until a new message has successfully staged
 * for transmission.
 */
  cs->state = CCB_CONTROL_SERVER_IDLE;
/*
 * Encode the message in the network buffer.
 */
  if(ccb_obj_to_net(&ccb_control_reply_table, nss_get_NetBuf(cs->nss),
		    &head->base)) {
    ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(12),
	"cs_stage_reply: ccb_obj_to_net() error (%s).\n", strerror(errno));
    return 1;
  };
/*
 * Mark the control link as busy.
 */
  cs->state = CCB_CONTROL_SERVER_BUSY;
  return 0;
}

/*.......................................................................
 * Query the current status of the control server.
 *
 * Input:
 *  cs           CCBControlServer *  The control server.
 * Output:
 *  return CCBControlServerStatus    The current status of the control
 *                                   server.
 */
CCBControlServerStatus ccb_cs_status(CCBControlServer *cs)
{
  return cs ? cs->state : CCB_CONTROL_SERVER_DOWN;
}

/*.......................................................................
 * Attempt to send queued control messages.
 *
 * Input:
 *  cs   CCBControlServer *   The control server.
 * Output:
 *  return              int     0 - OK.
 *                              1 - Error.
 */
int ccb_cs_send_control(CCBControlServer *cs)
{
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(cs ? cs->server : NULL, CCB_ERROR, CCB_LOGID(13),
		"ccb_cs_send_control: NULL argument(s).\n");
    return 1;
  };
/*
 * Keep writing until either all messages have been sent, or the output
 * socket blocks.
 */
  while(cs->state == CCB_CONTROL_SERVER_BUSY) {
/*
 * Attempt to write the latest message to the network.
 */
    switch(nss_send_msg(cs->nss)) {
    case NET_SEND_DATA:     /* Output temporarily blocked */
      return 0;
      break;
    case NET_SEND_DONE:     /* Message written, stage the next message */
/*
 * Mark the control server as idle until we have a new message to be sent.
 */
      cs->state = CCB_CONTROL_SERVER_IDLE; /* No more messages to send */
/*
 * If possible, get a new message from the queue and stage it for
 * transmission. Note that if a new message is staged, cs->state
 * will change to CCB_CONTROL_SERVER_BUSY.
 */
      if(cs_queue_control_reply(cs, NULL))
	return 1;
      break;
    case NET_SEND_CLOSED:
      ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(14),
	  "The CCB server attempted to write to a closed control socket.\n");
      return 1;
      break;
    case NET_SEND_ERROR:
    default:
      ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(15),
   "The CCB server encountered an I/O error while sending a control reply.\n");
      return 1;
      break;
    };
  };
/*
 * To get here, all queued messages must now have been sent.
 */
  return 0;
}

/*.......................................................................
 * Detach the control server from the control socket.
 *
 * Input:
 *  cs   CCBControlServer *   The control server.
 */
void ccb_cs_detach(CCBControlServer *cs)
{
/*
 * Detach from the current control socket.
 */
  detach_NetSendStr(cs->nss);
  detach_NetReadStr(cs->nrs);
/*
 * Mark the server as down.
 */
  cs->state = CCB_CONTROL_SERVER_DOWN;
  return;
}

/*.......................................................................
 * Attach the control server to a new socket.
 *
 * Input:
 *  cs   CCBControlServer *   The control server.
 *  fd                int     The file descriptor of the new control
 *                            socket.
 * Output:
 *  return                    0 - OK.
 *                            1 - Error.
 */
int ccb_cs_attach(CCBControlServer *cs, int fd)
{
/*
 * Check the arguments.
 */
  if(!cs || fd < 0) {
    if(!cs) {
      ccb_log_server_msg(cs ? cs->server : NULL, CCB_ERROR, CCB_LOGID(16),
		  "ccb_cs_attach: NULL argument(s).\n");
    } else {
      ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(17),
		  "ccb_cs_attach: Invalid socket.\n");
    };
    return 1;
  };
/*
 * Detach from any existing socket.
 */
  ccb_cs_detach(cs);
/*
 * Attach the output stream to the new socket.
 */
  attach_NetSendStr(cs->nss, fd);
  attach_NetReadStr(cs->nrs, fd);
/*
 * Mark the connection as idle.
 */
  cs->state = CCB_CONTROL_SERVER_IDLE;
/*
 * Stage for transmission the highest priority of any messages that
 * remain in any of the queues.
 */
  if(cs_queue_control_reply(cs, NULL))
    return 1;
  return 0;
}

/*.......................................................................
 * Read up to one command message from the CCB server control socket, and
 * when a complete command-message has been read, deliver it to the
 * application by calling the appropriate callback function.
 *
 * Input:
 *  cs    CCBControlServer *  The control server.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_cs_read_control(CCBControlServer *cs)
{
  CCBControlCommand cmd;   /* A generic command-message structure */
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(cs ? cs->server : NULL, CCB_ERROR, CCB_LOGID(18),
		"ccb_cs_read_control: NULL argument(s).\n");
    return 1;
  };
/*
 * Do nothing if no manager is currently connected.
 */
  if(cs->state == CCB_CONTROL_SERVER_DOWN)
    return 0;
/*
 * Attempt to read more data from the control socket.
 */
  switch(nrs_read_msg(cs->nrs)) {
  case NET_READ_SIZE:
  case NET_READ_DATA:   /* Input blocked before a message could be read */
    break;
  case NET_READ_DONE:   /* A new message has now been completely read */
/*
 * Decode the message into a message structure.
 */
    if(ccb_net_to_obj(&ccb_control_cmd_table, nrs_get_NetBuf(cs->nrs),
		      &cmd.head.base)) {
      ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(19),
		  "ccb_cs_read_control: ccb_net_to_obj() error (%s)",
		  strerror(errno));
      return 1;
    };
/*
 * Invoke the appropriate message-delivery callback.
 */
    if(cs_deliver_command(cs, &cmd))
      return 1;
    break;
  case NET_READ_CLOSED: /* There is no control socket */
    ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(20),
	"The CCB server attempted to read from a closed control socket.\n");
    return 1;
    break;
  case NET_READ_ERROR:  /* I/O error */
  default:
    ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(21),
 "The CCB server encountered an I/O error while reading a control command.\n");
      return 1;
    break;
  };
  return 0;
}

/*.......................................................................
 * Invoke the appropriate callback to deliver a newly received control
 * command to the CCB server process.
 *
 * Input:
 *  cs    CCBControlServer *  The control server.
 *  cmd  CCBControlCommand *  The command to be delivered.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int cs_deliver_command(CCBControlServer *cs, CCBControlCommand *cmd)
{
  CCBCmdStatus status;  /* The return status of the command */
/*
 * Be optimistic about the completion status of the command until proven
 * otherwise.
 */
  status = CCB_CMD_ACCEPTED;
/*
 * Respond to the received command.
 */
  switch(cmd->head.base.type) {
  case CCB_PHASE_SWITCH_CMD: /* A phase-switch configuration command */
    {
      CCBPhaseSwitchCnf *c = &cmd->phase_cmd.cnf;
      if(ccb_set_phase_switch_cnf(cs->cnf, c->active_switches,
				  c->closed_switches, c->samp_per_state))
	status = CCB_CMD_GARBLED;
    };
    break;
  case CCB_CAL_DIODE_CMD:    /* A cal-diode configuration command */
    {
      CCBCalDiodeCnf *c = &cmd->diode_cmd.cnf;
      if(ccb_set_cal_diode_cnf(cs->cnf, c->ncal, c->diode_states,
			       c->diode_times))
	status = CCB_CMD_GARBLED;
    };
    break;
  case CCB_TIMING_CMD:       /* A timing configuration command */
    {
      CCBTimingCnf *c = &cmd->timing_cmd.cnf;
      if(ccb_set_timing_cnf(cs->cnf, c->phase_switch_dt, c->diode_rise_dt,
			    c->diode_fall_dt, c->integ_period, c->roundtrip_dt,
			    c->holdoff_dt, c->adc_delay_dt))
	status = CCB_CMD_GARBLED;
    };
    break;
  case CCB_SAMPLER_CMD:       /* A sampler configuration command */
    {
      CCBSamplerCnf *c = &cmd->sampler_cmd.cnf;
      if(ccb_set_sampler_cnf(cs->cnf, c->sample_type))
	status = CCB_CMD_GARBLED;
    };
    break;
  case CCB_START_SCAN_CMD:   /* A start-scan command */
    if(ccb_cs_stage_scan(cs, cmd->start_scan.scan, cmd->start_scan.mjd,
			 cmd->start_scan.tod))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_STOP_SCAN_CMD:    /* A stop-scan command */
    if(ccb_cs_intra_scan(cs, cmd->stop_scan.scan))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_DUMP_SCAN_CMD:    /* A dump-scan command */
    if(ccb_cs_dump_scan(cs, cmd->dump_scan.scan, cmd->dump_scan.adc,
			cmd->dump_scan.samples, cmd->dump_scan.frames))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_RESET_CMD:        /* A reset command */
    if(ccb_reset_driver(cs->server))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_PING_CMD:    /* a ping command */
/*
 * Queue ping replies over both of the links.
 */
    if(cs_queue_cntrl_ping_reply(cs, cmd->ping.seq) ||
       ccb_queue_telem_ping_reply(cs->server, cmd->ping.seq))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_STATUS_REQUEST_CMD: /* a status-request command */
    if(cs_queue_status_reply(cs, ccb_general_status(cs->server)))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_SHUTDOWN_CMD:     /* A computer shutdown command */
    if(ccb_shutdown_rtc(cs->server))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_REBOOT_CMD:       /* A computer reboot command */
    if(ccb_reboot_rtc(cs->server))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_TELEMETRY_CMD:    /* A telemetry-stream control command */
    if(ccb_select_streams(cs->server, cmd->telemetry.streams))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_MONITOR_CMD:    /* A monitoring control command */
    if(ccb_monitor_period(cs->server, cmd->monitor.period))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_LOGGER_CMD:    /* A log control command */
    if(ccb_log_msg_lifetime(cs->server, cmd->logger.period))
      status = CCB_CMD_SYSERR;
    break;
  case CCB_LOAD_DRIVER_CMD:   /* A driver-selection command */
    if(ccb_load_driver(cs->server, cmd->driver.type))
      status = CCB_CMD_SYSERR;
    break;
  default:
    ccb_log_server_msg(cs->server, CCB_ERROR, CCB_LOGID(22),
      "cs_deliver_command: Can't deliver unknown type of command.\n");
    status = CCB_CMD_SYSERR;
    break;
  };
/*
 * Queue a command-acknowledgement reply to be sent to the manager.
 */
  return cs_queue_cmd_ack(cs, cmd->head.id, status);
}

/*.......................................................................
 * Reset the control server to its power-on defaults.
 *
 * Input:
 *  cs    CCBControlServer *  The control server.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_cs_reset(CCBControlServer *cs)
{
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(cs ? cs->server : NULL, CCB_ERROR, CCB_LOGID(23),
		"ccb_cs_reset: NULL argument(s).\n");
    return 1;
  };
/*
 * Reset the configuration parameters to their power-on defaults.
 */
  ccb_default_config(cs->cnf);
/*
 * Clear legacy messages from all queues.
 */
  ccb_mq_purge(cs->rq);
  return 0;
}

/*.......................................................................
 * Install the current CCB configuration parameters, and start an
 * intra-scan.
 *
 * Input:
 *  cs    CCBControlServer *  The control server.
 *  scan     unsigned long    The numeric identifier of the new scan.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_cs_intra_scan(CCBControlServer *cs, unsigned long scan)
{
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(cs ? cs->server : NULL, CCB_ERROR, CCB_LOGID(24),
		"ccb_cs_intra_scan: NULL argument(s).\n");
    return 1;
  };
/*
 * Install the latest configuration parameters, then start a new intra-scan.
 */
  if(ccb_driver_conf_scan(cs->server, cs->cnf) ||
     ccb_driver_intra_scan(cs->server, scan))
    return 1;
  return 0;
}

/*.......................................................................
 * Install the current CCB configuration parameters, and start a
 * dump-mode scan.
 *
 * Input:
 *  cs    CCBControlServer *  The control server.
 *  scan     unsigned long    The numeric ID to give the new intra-scan.
 *  adc     unsigned short    The ID of the ADC whose samples are to be
 *                            collected (0-15).
 *  samples  unsigned long    The max number of samples to collect per
 *                            integration.
 *  frames   unsigned long    The number of per-integration data-frames
 *                            to deliver.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_cs_dump_scan(CCBControlServer *cs, unsigned long scan,
		     unsigned short adc, unsigned long samples,
		     unsigned long frames)
{
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(cs ? cs->server : NULL, CCB_ERROR, CCB_LOGID(25),
		"ccb_cs_intra_scan: NULL argument(s).\n");
    return 1;
  };
/*
 * Install the latest configuration parameters, then start a new dump-mode
 * scan.
 */
  if(ccb_driver_conf_scan(cs->server, cs->cnf) ||
     ccb_driver_dump_scan(cs->server, scan, adc, samples, frames))
    return 1;
  return 0;
}

/*.......................................................................
 * Install the current CCB configuration parameters, and request that
 * a new scan be started at a given time.
 *
 * Input:
 *  cs    CCBControlServer *  The control server.
 *  scan     unsigned long    The numeric identifier of the new scan.
 *  mjd      unsigned long    The day on which to start the scan,
 *                            expressed as a Modified Julian Day number.
 *  secs     unsigned long    The time-of-day at which to start the scan,
 *                            expressed as the number of seconds from the
 *                            start of the specified day.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
static int ccb_cs_stage_scan(CCBControlServer *cs, unsigned long scan,
			     unsigned long mjd, unsigned long secs)
{
/*
 * Check the arguments.
 */
  if(!cs) {
    ccb_log_server_msg(cs ? cs->server : NULL, CCB_ERROR, CCB_LOGID(26),
		"ccb_cs_stage_scan: NULL argument(s).\n");
    return 1;
  };
/*
 * Install the latest configuration parameters, then stage a new scan
 * to be started at the specified time.
 */
  if(ccb_driver_conf_scan(cs->server, cs->cnf) ||
     ccb_driver_stage_scan(cs->server, scan, mjd, secs))
    return 1;
  return 0;
}
