/*
 * Standard C-library includes.
 */
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

/*
 * Local includes.
 */
#include "ccbconstants.h"
#include "ccbserverlink.h"
#include "ccbrealdriver.h"

/*
 * Encapsulate the running parameters of the demo program. Note that a
 * typedef alias for this struct is defined in ccbdemodriver.c.
 */
struct CCBRealDriverData {
  int place_holder;
};

static CCB_DRIVER_LOAD_FN(real_driver_load_fn);
static CCB_DRIVER_UNLOAD_FN(real_driver_unload_fn);
static CCB_DRIVER_TELL_FN(real_driver_tell_fn);
static CCB_REBOOT_RTC_FN(real_reboot_rtc_fn);
static CCB_SHUTDOWN_RTC_FN(real_shutdown_rtc_fn);
static CCB_DRIVER_SELECT_EVENTS_FN(real_driver_select_events_fn);
static CCB_DRIVER_CHECK_EVENTS_FN(real_driver_check_events_fn);

static int function_not_implemented_yet(CCBServerLink *sl,
					CCBRealDriverData *dd);

/*.......................................................................
 * Create a new CCBRealDriverData object. This contains all of the resources
 * needed by the real CCB driver.
 *
 * Output:
 *  return  CCBRealDriverData *  The new object, or NULL on error.
 */
CCBRealDriverData *new_CCBRealDriverData(void)
{
  CCBRealDriverData *dd;  /* The object to be returned */
/*
 * Allocate the container.
 */
  dd = malloc(sizeof(CCBRealDriverData));
  if(!dd) {
    fprintf(stderr, "new_CCBRealDriverData: Insufficient memory.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBRealDriverData().
 */

  return dd;
}

/*.......................................................................
 * Delete a CCBRealDriverData object.
 *
 * Input:
 *  data    CCBRealDriverData *  The object to be deleted.
 * Output:
 *  return  CCBRealDriverData *  The deleted object (always NULL).
 */
CCBRealDriverData *del_CCBRealDriverData(CCBRealDriverData *dd)
{
  if(dd) {
/*
 * Delete the container of the object.
 */
    free(dd);
  };
  return NULL;
}

/*.......................................................................
 * Allocate the interface object of the real-driver.
 *
 * Input:
 *  data  CCBRealDriverData *  The resource object of the driver.
 * Output:
 *  return  CCBServerDriver *  The new interface object, or NULL on error.
 */
CCBServerDriver *new_CCBRealDriver(CCBRealDriverData *dd)
{
  return new_CCBServerDriver(dd, real_driver_load_fn,
			     real_driver_unload_fn, real_driver_tell_fn,
			     real_driver_select_events_fn,
			     real_driver_check_events_fn,
			     real_reboot_rtc_fn, real_shutdown_rtc_fn);
}

/*.......................................................................
 * Load the simulated CCB device driver.
 */
static CCB_DRIVER_LOAD_FN(real_driver_load_fn)
{
  CCBRealDriverData *dd = (CCBRealDriverData *) data;
  return function_not_implemented_yet(sl, dd);
}

/*.......................................................................
 * Unload the simulated CCB device driver.
 */
static CCB_DRIVER_UNLOAD_FN(real_driver_unload_fn)
{
  CCBRealDriverData *dd = (CCBRealDriverData *) data;
  return function_not_implemented_yet(sl, dd);
}

/*.......................................................................
 * Execute a device-driver command.
 */
static CCB_DRIVER_TELL_FN(real_driver_tell_fn)
{
  CCBRealDriverData *dd = (CCBRealDriverData *) data;
  return function_not_implemented_yet(sl, dd);
}

/*.......................................................................
 * Stage a simulated reboot.
 */
static CCB_REBOOT_RTC_FN(real_reboot_rtc_fn)
{
  CCBRealDriverData *dd = (CCBRealDriverData *) data;
  return function_not_implemented_yet(sl, dd);
}

/*.......................................................................
 * Stage a simulated shutdown.
 */
static CCB_SHUTDOWN_RTC_FN(real_shutdown_rtc_fn)
{
  CCBRealDriverData *dd = (CCBRealDriverData *) data;
  return function_not_implemented_yet(sl, dd);
}

/*.......................................................................
 * Update the specified select() arguments to register the real inactivity
 * timeout. This timeout is used to provide the simulator clock.
 */
static CCB_DRIVER_SELECT_EVENTS_FN(real_driver_select_events_fn)
{
  CCBRealDriverData *dd = (CCBRealDriverData *) data;
  return function_not_implemented_yet(sl, dd);
}

/*.......................................................................
 * Respond to a select() inactivity timeout.
 */
static CCB_DRIVER_CHECK_EVENTS_FN(real_driver_check_events_fn)
{
  CCBRealDriverData *dd = (CCBRealDriverData *) data;
  return function_not_implemented_yet(sl, dd);
}

/*.......................................................................
 * This is a temporary function called by unimplemented interface
 * functions. It reports an error to the server, and returns the normal
 * error-return code of the interface callback-functions.
 */
static int function_not_implemented_yet(CCBServerLink *sl,
					CCBRealDriverData *dd)
{
  ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(211),
		     "Unimplemented function called.\n");
  ccb_record_server_faults(sl, CCB_SOFT_FAULT);
  return 1;
}
