/* Standard C includes */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdarg.h>

/* Socket includes */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/* Local includes */

#include "ccbserverlink.h"
#include "ccbconstants.h"
#include "ccbtelemetryserver.h"
#include "ccbcontrolserver.h"
#include "ccbprint.h"
#include "iptable.h"
#include "ccbconfdir.h"
#include "tcpip.h"
#include "ccbserverpriv.h"

/*
 * Encapsulate callback functions and associated callback data in
 * a structure of the following type. This struct is typedef'd as
 * CCBServerDriver in ccbserverlink.h.
 */
struct CCBServerDriver {
  void *data;                             /* App-specific callback data */
  CCBDriverLoadFn *load_driver;           /* Load the device-driver */
  CCBDriverUnloadFn *unload_driver;       /* Unload the device-driver */
  CCBDriverTellFn *tell_driver;           /* Command the device-driver */
  CCBDriverSelectEventsFn *select_events; /* Register the driver's select() */
                                          /*  event-watching needs. */
  CCBDriverCheckEventsFn *check_events;   /* Respond to select() activity */
                                          /*  reports. */
  CCBRebootRTCFn *reboot_rtc;             /* Reboot the CCB real-time cpu */
  CCBShutdownRTCFn *shutdown_rtc;         /* Shutdown the CCB real-time cpu */
};

/*
 * The following type of object, which is given the typedef CCBServerLink alias
 * in ccbserverlink.h, holds all of the resources used to manage communications
 * between the calling CCB server and a remote CCB manager.
 */
struct CCBServerLink {
  int cntrl_port;          /* The socket bound to the server's control port */
  int telem_port;          /* The socket bound to the server's telemetry */
                           /*  port. */
  int cntrl_sock;          /* The current control socket */
  int telem_sock;          /* The current telemetry socket */
  CCBTelemetryServer *ts;  /* The telemetry server */
  CCBControlServer *cs;    /* The control server */
  int log_recursion_count; /* The number of times that */
                           /* ccb_log_server_msg() is being recursively */
                           /* invoked. */
  CCBServerDriver *normal; /* Methods and data for the real CCB device driver */
  CCBServerDriver *virtual;/* Methods and data for the fake CCB device driver */
  CCBServerDriver *driver; /* Either sl->normal or sl->virtual, depending on */
                           /*  which is currently selected. */
  CCBIPTable *iptable;     /* The table of valid authorized addresses */
  unsigned faults;         /* Fault conditions reported to */
                           /*  ccb_record_server_faults() */
};

static int ccb_server_error(CCBServerLink *sl);
static int ccb_read_control_port(CCBServerLink *sl);
static int ccb_read_telemetry_port(CCBServerLink *sl);
static int ccb_driver_not_loaded(CCBServerLink *sl);

/*
 * Enumerate the possible types of CCB-server socket I/O to watch for.
 */
typedef enum {
  CCB_RD_CNTRL_SOCK=1,   /* Watch for control socket readability */
  CCB_WR_CNTRL_SOCK=2,   /* Watch for control socket writability */
  CCB_RD_TELEM_SOCK=4,  /* Watch for telemetry socket readability */
  CCB_WR_TELEM_SOCK=8,  /* Watch for telemetry socket writability */
  CCB_RD_CNTRL_PORT=16,  /* Watch for control port connection */
  CCB_RD_TELEM_PORT=32  /* Watch for telemetry-port connections */
} CCBServerIOStatus;

/*
 * Convert the contents of select() file-descriptor sets to a
 * set an I/O mask of the form expected by ccb_server_communicate().
 */
static unsigned ccb_server_select_to_io(CCBServerLink *sl, const fd_set *rfds,
					const fd_set *wfds);

/*
 * Perform socket-level I/O on the CCB server sockets.
 */
static int ccb_server_communicate(CCBServerLink *ccb, const fd_set *rfds,
				  const fd_set *wfds);

/*
 * Augment select() file-descriptor sets according to the socket
 * I/O that the server library is watching for.
 */
static int ccb_server_select_args(CCBServerLink *sl, fd_set *rfds, fd_set *wfds,
				  int *maxfd);

/*
 * Return a bitwise union of CCBServerIOStatus enumerators to
 * list the types of I/O that the server should be watching for.
 */
static unsigned ccb_server_io_status(CCBServerLink *sl);

/*.......................................................................
 * Create a new CCBServerLink object.
 *
 * Input:
 *  normal  CCBServerDriver *  The encapsulated method functions of the
 *                             real CCB device driver. This can be NULL.
 *  virtual CCBServerDriver *  The encapsulated method functions of an
 *                             emulation of the CCB device driver and its
 *                             hardware.
 * Output:
 *  return    CCBServerLink *  The new object, or NULL on error.
 */
CCBServerLink *new_CCBServerLink(CCBServerDriver *normal,
				 CCBServerDriver *virtual)
{
  CCBServerLink *sl;    /* The object to be returned */
  char errmsg[200];     /* An error message buffer */
  const char *confdir;  /* The CCB configuration directory */
/*
 * Check the arguments.
 */
  if(!virtual) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(28),
	"new_CCBServerLink: NULL argument(s)\n");
    return NULL;
  };
/*
 * Allocate the container.
 */
  sl = malloc(sizeof(CCBServerLink));
  if(!sl) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(29),
	"Insufficient memory for CCB server communications resources.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBServerLink().
 */
  sl->cntrl_port = -1;
  sl->telem_port = -1;
  sl->cntrl_sock = -1;
  sl->telem_sock = -1;
  sl->ts = NULL;
  sl->cs = NULL;
  sl->log_recursion_count = 0;
  sl->normal = normal;
  sl->virtual = virtual;
  sl->driver = NULL;     /* Start with no device driver */
  sl->iptable = NULL;
  sl->faults = 0;
/*
 * Bind the telemetry and control ports of the server.
 */
  if((sl->cntrl_port = tcp_server_sock(CCB_CONTROL_PORT,
					  CCB_SERVER_BACKLOG)) < 0) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(30),
		"Error binding CCB control port (%s)", strerror(errno));
    return del_CCBServerLink(sl);
  };
  if((sl->telem_port = tcp_server_sock(CCB_TELEMETRY_PORT,
					   CCB_SERVER_BACKLOG)) < 0) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(31),
		"Error binding CCB telemetry port (%s)", strerror(errno));
    return del_CCBServerLink(sl);
  };
/*
 * Set the server ports to non-blocking mode, so that calls to accept()
 * don't block.
 */
  if(tcp_set_blocking(sl->cntrl_port, 0) ||
     tcp_set_blocking(sl->telem_port, 0)) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(32),
		"Error switching ports to non-blocking (%s).\n",
		strerror(errno));
    return del_CCBServerLink(sl);
  };
/*
 * Allocate the telemetry server object.
 */
  sl->ts = new_CCBTelemetryServer(sl);
  if(!sl->ts)
    return del_CCBServerLink(sl);
/*
 * Allocate the control server object.
 */
  sl->cs = new_CCBControlServer(sl);
  if(!sl->cs)
    return del_CCBServerLink(sl);
/*
 * Get the CCB configuration directory. If the run-time CCB_CONF_DIR
 * environment variable is defined, interpret its contents as the path
 * of the configuration directory. Otherwise use the contents of the
 * compile-time CCB_CONF_DIR macro (configurable via the
 * --sysconfdir=DIR argument of the configure script).
 */
  confdir = getenv("CCB_CONF_DIR");
  if(!confdir || *confdir=='\0')
    confdir = CCB_CONF_DIR;
/*
 * Read the table of authorized IP addresses.
 */
  sl->iptable = new_CCBIPTable(confdir, CCB_IP_TABLE_FILE,
			       errmsg, sizeof(errmsg));
  if(!sl->iptable) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(33), "%s", errmsg);
    return del_CCBServerLink(sl);
  };
/*
 * The server resources are now ready for use.
 */
  return sl;
}

/*.......................................................................
 * Delete a CCBServerLink object.
 *
 * Input:
 *  sl       CCBServerLink *  The object to be deleted.
 * Output:
 *  return   CCBServerLink *  The deleted object (always NULL).
 */
CCBServerLink *del_CCBServerLink(CCBServerLink *sl)
{
  if(sl) {
/*
 * Delete the resources of the telemetry and control servers.
 */
    sl->ts = del_CCBTelemetryServer(sl->ts);
    sl->cs = del_CCBControlServer(sl->cs);
/*
 * If a manager is currently connected, close the control and
 * telemetry sockets that were being used to communicate with
 * it.
 */
    if(sl->cntrl_sock >= 0)
      close(sl->cntrl_sock);
    if(sl->telem_sock >= 0)
      close(sl->telem_sock);
/*
 * Close the server ports.
 */
    if(sl->cntrl_port >= 0)
      close(sl->cntrl_port);
    if(sl->telem_port >= 0)
      close(sl->telem_port);
/*
 * Delete the list of authorized manager IP addresses.
 */
    sl->iptable = del_CCBIPTable(sl->iptable);
/*
 * Delete the now-empty container.
 */
    free(sl);
  };
  return NULL;
}

/*.......................................................................
 * Create a new CCBServerDriver object.
 *
 * Input:
 *  data                             void *  Anonymous application-specific
 *                                           data to be passed to the
 *                                           following callback functions.
 *  load_driver           CCBDriverLoadFn *  The function to call to load
 *                                           the CCB device-driver.
 *  unload_driver       CCBDriverUnloadFn *  The function to call to unload
 *                                           the CCB device-driver.
 *  tell_driver           CCBDriverTellFn *  The function to call to send
 *                                           the CCB device-driver a command.
 *  select_events CCBDriverSelectEventsFn *  Get the driver to tell us which
 *                                           select() events to watch.
 *  check_events   CCBDriverCheckEventsFn *  Get the driver to respond to an
 *                                           event reported by select().
 *  reboot_rtc             CCBRebootRTCFn *  The function to call to reboot
 *                                           the CCB real-time computer.
 *  shutdown_rtc         CCBShutdownRTCFn *  The function to call to shutdown
 *                                           the CCB real-time computer.
 * Output:
 *  return                CCBServerDriver *  The new object, or NULL on error.
 */
CCBServerDriver *new_CCBServerDriver(void *data,
				     CCBDriverLoadFn *load_driver,
				     CCBDriverUnloadFn *unload_driver,
				     CCBDriverTellFn *tell_driver,
				     CCBDriverSelectEventsFn *select_events,
				     CCBDriverCheckEventsFn *check_events,
				     CCBRebootRTCFn *reboot_rtc,
				     CCBShutdownRTCFn *shutdown_rtc)
{
  CCBServerDriver *drv;  /* The object to be returned */
/*
 * Check the arguments.
 */
  if(!load_driver || !unload_driver || !tell_driver || !select_events ||
     !check_events || !reboot_rtc || !shutdown_rtc) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(34),
	"new_CCBServerDriver: NULL argument(s)\n");
    return NULL;
  };
/*
 * Allocate the container.
 */
  drv = malloc(sizeof(CCBServerDriver));
  if(!drv) {
    ccb_log_server_msg(NULL, CCB_ERROR, CCB_LOGID(35),
	"Insufficient memory for CCB server communications resources.\n");
    return NULL;
  };
/*
 * Before attempting any operation that might fail, initialize the
 * container at least up to the point at which it can safely be passed
 * to del_CCBServerDriver().
 */
  drv->data = data;
  drv->load_driver = load_driver;
  drv->unload_driver = unload_driver;
  drv->tell_driver = tell_driver;
  drv->select_events = select_events;
  drv->check_events = check_events;
  drv->reboot_rtc = reboot_rtc;
  drv->shutdown_rtc = shutdown_rtc;
/*
 * The server resources are now ready for use.
 */
  return drv;
}

/*.......................................................................
 * Delete a CCBServerDriver object.
 *
 * Input:
 *  drv      CCBServerDriver *  The object to be deleted.
 * Output:
 *  return   CCBServerDriver *  The deleted object (always NULL).
 */
CCBServerDriver *del_CCBServerDriver(CCBServerDriver *drv)
{
  if(drv) {
/*
 * Delete the now-empty container.
 */
    free(drv);
  };
  return NULL;
}

/*.......................................................................
 * This is a printf-style function that queues a formatted log message for
 * subsequent dispatch to the manager.
 *
 * Input:
 *  sl  CCBServerLink *  The communications resource object of the CCB
 *                       server, or NULL to write to stderr.
 *  level CCBLogLevel    The significance of the message.
 *  id  unsigned long    The numeric identifier of the message. This must
 *                       be chosen to be unique within the CCB software,
 *                       and be the lowest so-far unused id number >= 0.
 *  fmt    const char *  This is a standard printf-style format string.
 *  ...                  The arguments refered to by the format specifiers
 *                       in the fmt[] argument.
 * Output:
 *  return        int    0 - OK.
 *                       1 - Error.
 */
int ccb_log_server_msg(CCBServerLink *sl, CCBLogLevel level, unsigned long id,
		       const char *fmt, ...)
{
  int status;               /* The return value of ccb_vsnprintf() */
  int retval = 0;           /* The return value of this function */
  char buffer[CCB_MAX_LOG]; /* The buffer in which the message is formatted */
  va_list ap;               /* The encapsulated variable argument list */
/*
 * Check the arguments.
 */
  if(!fmt) {
    fmt = "ccb_log_server_msg: Missing log-message format string.";
    level = CCB_ERROR;
    retval = 1;
  };
/*
 * Increment the count of the number of recursive calls to this function
 * that are currently taking place. This allows us to avoid an infinite
 * recursion if any of the functions that this function calls, themselves
 * call this function.
 */
  if(sl)
    sl->log_recursion_count++;
/*
 * Compose the message.
 */
  va_start(ap, fmt);
  status = ccb_vsnprintf(buffer, sizeof(buffer), fmt, ap);
  va_end(ap);
/*
 * Error formatting message?
 */
  if(status < 0) {
    strcpy(buffer, "Error formatting log message.");
    level = CCB_ERROR;
    retval = 1;
/*
 * If the message contains any newline characters, terminate the string
 * at the first one found.
 */
  } else {
    char *cptr = strchr(buffer, '\n');
    if(cptr)
      *cptr = '\0';
  };
/*
 * If a sl object has been provided, and this function isn't being
 * called recursively, queue the message to be sent to the
 * manager. Otherwise write it to stderr.
 */
  if(sl && sl->log_recursion_count == 1) {
    if(ccb_ts_queue_log_msg(sl->ts, level, id, buffer))
      retval = 1;
  } else {
    (void) fprintf(stderr, "%s\n", buffer);
  };
/*
 * Decrement the recursion count of the function, now that we are leaving it.
 */
  if(sl)
    sl->log_recursion_count--;
  return retval;
}

/*.......................................................................
 * Queue a new integration message.
 *
 * Input:
 *  sl            CCBServerLink *  The CCB server.
 *  t        const CCBTimeStamp *  The time at which the integration was
 *                                 recorded.
 *  scan          unsigned long    The index of the scan that originated
 *                                 the integration.
 *  number        unsigned long    The sequential number of the integration
 *                                 within the scan.
 *  flags        unsigned short    A bitwise union of CCBIntegFlags enumerators,
 *                                 describing the status of the hardware during
 *                                 the integration.
 *  values  const unsigned long *  The integrated values.
 *  nvalues       unsigned long    The number of values in values[]. This
 *                                 should be CCB_MAX_INTEG. If less than
 *                                 this the remaining values will be set
 *                                 to zero, and if more are passed, the
 *                                 extra elements will be ignored.
 * Output:
 *  return                  int    0 - OK.
 *                                 1 - Error.
 */
int ccb_queue_integ_msg(CCBServerLink *sl, const CCBTimeStamp *t,
			unsigned long scan, unsigned long number,
			unsigned short flags, const unsigned long *values,
			unsigned nvalues)
{
/*
 * Check the arguments.
 */
  if(!sl || !t || !values) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(36),
	"ccb_queue_integ_msg: NULL argument(s).\n");    
    return 1;
  };
/*
 * Forward the request to the telemetry server.
 */
  return ccb_ts_queue_integ_msg(sl->ts, t, scan, number, flags, values,
				nvalues);
}

/*.......................................................................
 * Queue a new monitor message.
 *
 * Input:
 *  sl             CCBServerLink *  The CCB server.
 *  t         const CCBTimeStamp *  The time at which the monitor was
 *                                  recorded.
 *  scan           unsigned long    The sequential number of the parent scan.
 *  number         unsigned long    The sequential number of the monitor.
 *  values         unsigned long *  The integrated values.
 *  nvalues  const unsigned long    The number of values in values[]. This
 *                                  should be CCB_MAX_MONITOR. If less than
 *                                  this, the remaining values will be set
 *                                  to zero, and if more are passed, the
 *                                  extra elements will be ignored.
 * Output:
 *  return                   int    0 - OK.
 *                                  1 - Error.
 */
int ccb_queue_monitor_msg(CCBServerLink *sl, const CCBTimeStamp *t,
			  unsigned long scan, unsigned long number,
			  const unsigned long *values, unsigned nvalues)
{
/*
 * Check the arguments.
 */
  if(!sl || !t || !values) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(37),
	"ccb_queue_monitor_msg: NULL argument(s).\n");    
    return 1;
  };
/*
 * Forward the request to the telemetry server.
 */
  return ccb_ts_queue_monitor_msg(sl->ts, t, scan, number, values, nvalues);
}

/*.......................................................................
 * Queue a ping reply for subsequent transmission to the manager.
 *
 * Input:
 *  sl   CCBServerLink *   The CCB server.
 *  seq  unsigned long     The sequence number of the ping command.
 * Output:
 *  return         int     0 - OK.
 *                         1 - Error.
 */
int ccb_queue_telem_ping_reply(CCBServerLink *sl, unsigned long seq)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(38),
	"ccb_queue_telem_ping_reply: NULL argument(s).\n");    
    return 1;
  };
/*
 * Forward the request to the telemetry server.
 */
  return ccb_ts_queue_telem_ping_reply(sl->ts, seq);
}

/*.......................................................................
 * Perform socket-level I/O on the CCB server sockets.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 *  rfds   const fd_set *  The set of readable file-descriptors returned
 *                         by select().
 *  wfds   const fd_set *  The set of writable file-descriptors returned
 *                         by select().
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
static int ccb_server_communicate(CCBServerLink *sl, const fd_set *rfds,
				  const fd_set *wfds)
{
  unsigned io;  /* The set of I/O events to perform */
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(39),
	"ccb_server_communicate: NULL argument(s).\n");    
    return 1;
  };
/*
 * Determine which types of I/O select() reported as now being possible.
 */
  io = ccb_server_select_to_io(sl, rfds, wfds);
/*
 * Remove types of I/O that aren't currently expected.
 */
  io &= ccb_server_io_status(sl);
/*
 * Read from the control socket?
 */
  if((io & CCB_RD_CNTRL_SOCK && ccb_cs_read_control(sl->cs)) ||
/*
 * Write to the control socket?
 */
     (io & CCB_WR_CNTRL_SOCK && ccb_cs_send_control(sl->cs)) ||
/*
 * Read from the telemetry socket?
 */
     (io & CCB_RD_TELEM_SOCK && ccb_ts_read_telemetry(sl->ts)) ||
/*
 * Write to the telemetry socket?
 */
     (io & CCB_WR_TELEM_SOCK && ccb_ts_send_telemetry(sl->ts)) ||
/*
 * Has a connection request arrived over the control port?
 */
     (io & CCB_RD_CNTRL_PORT && ccb_read_control_port(sl)) ||
/*
 * Has a connection request arrived over the telemetry port?
 */
     (io & CCB_RD_TELEM_PORT && ccb_read_telemetry_port(sl))) {
/*
 * If any of the above statements failed, terminate the connection.
 */
    ccb_server_error(sl);
  };
  return 0;
}

/*.......................................................................
 * Respond to a connection request on the control port.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
static int ccb_read_control_port(CCBServerLink *sl)
{
  struct sockaddr_in addr;      /* The IP address of the prospective client */
  socklen_t len = sizeof(addr); /* The number of bytes in addr */
  int fd;                       /* The socket assigned to the client */
  char buffer[INET_ADDRSTRLEN]; /* A buffer in which to decode the address */
/*
 * Tentatively accept the new client.
 */
  fd = accept(sl->cntrl_port, (struct sockaddr *) &addr, &len);
/*
 * Did the accept fail?
 */
  if(fd < 0) {
/*
 * Check errno to see why accept failed.
 */
    switch(errno) {
#ifdef EWOULDBLOCK       /* The client aborted the connection (BSD) */
    case EWOULDBLOCK:
#endif
#ifdef ECONNABORTED      /* The client aborted the connection (POSIX.1g) */
    case ECONNABORTED:
#endif
#ifdef EPROTO            /* The client aborted the connection (SVR4) */
    case EPROTO:
#endif
    case EINTR:          /* A signal was received */
      return 0;
      break;
    default:
      ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(40),
		  "CCB control-server accept() error (%s).\n",
		  strerror(errno));
      return 1;
      break;
    };
  };
/*
 * Convert the binary address in 'addr' to a textual representation
 * of the numeric IP address.
 */
  if(inet_ntop(AF_INET, &addr.sin_addr, buffer, sizeof(buffer)) == NULL)
    strcpy(buffer, "(?)");
/*
 * If a manager is already connected or the client's address isn't in
 * the list of allowed manager IP addresses, report the new
 * connection, then close it to reject it.
 */
  if(sl->cntrl_sock >= 0 || !ccb_valid_manager_address(sl->iptable,
							  buffer)) {
    ccb_log_server_msg(sl, CCB_NOTICE, CCB_LOGID(41),
		"CCB control-server rejected connection from %s\n",
		buffer);
    close(fd);
    return 0;
  };
/*
 * We have a new manager. Woo hoo!
 */
  ccb_log_server_msg(sl, CCB_NOTICE, CCB_LOGID(42),
	      "The CCB control-server is accepting a new connection from %s\n",
	      buffer);
/*
 * Configure the socket for non-blocking I/O.
 */
  if(tcp_set_blocking(fd, 0)) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(43),
	"Error switching control socket to non-blocking socket I/O.");
    close(fd);
    return 1;
  };
/*
 * Pass the new socket descriptor to the control server.
 */
  if(ccb_cs_attach(sl->cs, fd)) {
    close(fd);
    return 1;
  };
/*
 * Record the new control socket.
 */
  sl->cntrl_sock = fd;
  return 0;
}

/*.......................................................................
 * Respond to a connection request on the telemetry port.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
static int ccb_read_telemetry_port(CCBServerLink *sl)
{
  struct sockaddr_in addr;      /* The IP address of the prospective client */
  socklen_t len = sizeof(addr); /* The number of bytes in addr */
  int fd;                       /* The socket assigned to the client */
  char buffer[INET_ADDRSTRLEN]; /* A buffer in which to decode the address */
/*
 * Tentatively accept the new client.
 */
  fd = accept(sl->telem_port, (struct sockaddr *) &addr, &len);
/*
 * Did the accept fail?
 */
  if(fd < 0) {
/*
 * Check errno to see why accept failed.
 */
    switch(errno) {
#ifdef EWOULDBLOCK       /* The client aborted the connection (BSD) */
    case EWOULDBLOCK:
#endif
#ifdef ECONNABORTED      /* The client aborted the connection (POSIX.1g) */
    case ECONNABORTED:
#endif
#ifdef EPROTO            /* The client aborted the connection (SVR4) */
    case EPROTO:
#endif
    case EINTR:          /* A signal was received */
      return 0;
      break;
    default:
      ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(44),
		  "CCB telemetry-server accept() error (%s).\n",
		  strerror(errno));
      return 1;
      break;
    };
  };
/*
 * Convert the binary address in 'addr' to a textual representation
 * of the numeric IP address.
 */
  if(inet_ntop(AF_INET, &addr.sin_addr, buffer, sizeof(buffer)) == NULL)
    strcpy(buffer, "(?)");
/*
 * If a manager is already connected or the client's address isn't in
 * the list of allowed manager IP addresses, report the new
 * connection, then close it to reject it.
 */
  if(sl->telem_sock >= 0 || !ccb_valid_manager_address(sl->iptable,
							  buffer)) {
    ccb_log_server_msg(sl, CCB_NOTICE, CCB_LOGID(45),
		"CCB telemetry-server rejected connection from %s\n",
		buffer);
    close(fd);
    return 0;
  };
/*
 * We have a new telemetry connection.
 */
  ccb_log_server_msg(sl, CCB_NOTICE, CCB_LOGID(46),
      "The CCB telemetry-server is accepting a new connection from %s\n",
      buffer);
/*
 * Configure the socket for non-blocking I/O.
 */
  if(tcp_set_blocking(fd, 0)) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(47),
	"Error switching telemetry socket to non-blocking socket I/O.");
    close(fd);
    return 1;
  };
/*
 * Pass the new socket descriptor to the telemetry server.
 */
  if(ccb_ts_attach(sl->ts, fd)) {
    close(fd);
    return 1;
  };
/*
 * Record the new telemetry socket.
 */
  sl->telem_sock = fd;
  return 0;
}

/*.......................................................................
 * Close both the control and telemetry connections of the current
 * manager client in response to an error, then return non-zero to
 * allow this function to be used in return statements that need to
 * return a non-zero error code.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 * Output:
 *  return          int    Always 1 (ie. the error return code of most
 *                         functions in this module).
 */
static int ccb_server_error(CCBServerLink *sl)
{
/*
 * Disconnect the telemetry and control servers.
 */
  ccb_cs_detach(sl->cs);
  ccb_ts_detach(sl->ts);
/*
 * Close the control and telemetry sockets.
 */
  if(sl->cntrl_sock >= 0) {
    close(sl->cntrl_sock);
    sl->cntrl_sock = -1;
  };
  if(sl->telem_sock >= 0) {
    close(sl->telem_sock);
    sl->telem_sock = -1;
  };
/*
 * Queue a log message to be sent to the next manager that connects to
 * the telemetry server.
 */
  ccb_log_server_msg(sl, CCB_FATAL, CCB_LOGID(48),
    "The CCB server disconnected the manager in response to an error.\n");
/*
 * Unload the CCB device driver.
 */
  if(ccb_unload_driver(sl))
    return 1;
/*
 * Return the normal error code of most functions in this module.
 */
  return 1;
}

/*.......................................................................
 * Return a bitwise union of CCBServerIOStatus enumerators to
 * list the types of I/O that the server should be watching for.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 * Output:
 *  return     unsigned    A bitwise union of CCBServerIOStatus
 *                         enumerators.
 */
static unsigned ccb_server_io_status(CCBServerLink *sl)
{
  unsigned status = 0;   /* The return value */
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(49),
	"ccb_server_io_status: NULL argument(s).\n");    
    return status;
  };
/*
 * The control and telemetry ports are always to be watched for
 * incoming connection requests.
 */
  status |= CCB_RD_CNTRL_PORT | CCB_RD_TELEM_PORT;
/*
 * Do we have a control connection?
 */
  if(sl->cntrl_sock >= 0) {
/*
 * While the control link is active, the server should always be
 * watching both for the arrival of new commands and for disconnection
 * by the remote manager.
 */
    status |= CCB_RD_CNTRL_SOCK;
/*
 * If there are any incompletely sent, or queued outgoing control replies,
 * tell the caller to watch the writability of the control socket.
 */
    if(ccb_cs_status(sl->cs) == CCB_CONTROL_SERVER_BUSY)
      status |= CCB_WR_CNTRL_SOCK;
  };
/*
 * Do we have a telemetry connection?
 */
  if(sl->telem_sock >= 0) {
/*
 * While the telemetry link is active, the server should always be
 * watching for disconnection events.
 */
    status |= CCB_RD_TELEM_SOCK;
/*
 * If there are any incompletely sent, or queued outgoing telemetry
 * messages, tell the caller to watch the writability of the telemetry
 * socket.
 */
    if(ccb_ts_status(sl->ts) == CCB_TELEM_SERVER_BUSY)
      status |= CCB_WR_TELEM_SOCK;
  };
/*
 * Return the accumulated set of I/O events to watch for.
 */
  return status;
}

/*.......................................................................
 * Augment select() file-descriptor sets according to the socket
 * I/O that the server library is watching for.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 * Input/Output:
 *  rfds         fd_set *  The server file-descriptors that should be
 *                         watched for readability will be added to the
 *                         existing contents of this set.
 *  wfds         fd_set *  The server file-descriptors that should be
 *                         watched for writability will be added to the
 *                         existing contents of this set.
 *  maxfd           int *  The maximum of the file descriptors set in
 *                         the above sets, and the value of *maxfd
 *                         on entry to this function.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
static int ccb_server_select_args(CCBServerLink *sl, fd_set *rfds, fd_set *wfds,
				  int *maxfd)
{
  unsigned io;  /* The set of I/O events to be watched */
/*
 * Check the arguments.
 */
  if(!sl || !rfds || !wfds || !maxfd) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(50),
	"ccb_server_select_args: NULL argument(s).\n");    
    return 1;
  };
/*
 * Lookup the set of I/O events that are to be watched.
 */
  io = ccb_server_io_status(sl);
/*
 * Transpose the contents of the I/O set to the corresponding select()
 * file-descriptor sets.
 */
  if(sl->cntrl_sock >= 0) {
    if(io & CCB_RD_CNTRL_SOCK)
      FD_SET(sl->cntrl_sock, rfds);
    if(io & CCB_WR_CNTRL_SOCK)
      FD_SET(sl->cntrl_sock, wfds);
    if(sl->cntrl_sock > *maxfd)
      *maxfd = sl->cntrl_sock;
  };
  if(sl->telem_sock >= 0) {
    if(io & CCB_RD_TELEM_SOCK)
      FD_SET(sl->telem_sock, rfds);
    if(io & CCB_WR_TELEM_SOCK)
      FD_SET(sl->telem_sock, wfds);
    if(sl->telem_sock > *maxfd)
      *maxfd = sl->telem_sock;
  };
  if(io & CCB_RD_TELEM_PORT) {
    FD_SET(sl->telem_port, rfds);
    if(sl->telem_port > *maxfd)
      *maxfd = sl->telem_port;
  };
  if(io & CCB_RD_CNTRL_PORT) {
    FD_SET(sl->cntrl_port, rfds);
    if(sl->cntrl_port > *maxfd)
      *maxfd = sl->cntrl_port;
  };
  return 0;
}

/*.......................................................................
 * Convert the contents of select() file-descriptor sets to a
 * set an I/O mask of the form expected by ccb_server_communicate().
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 *  rfds   const fd_set *  The set of file-descriptors that should be
 *                         watched for readability.
 *  wfds   const fd_set *  The set of file-descriptors that should be
 *                         watched for writability.
 * Output:
 *  return     unsigned    The bitmask union of CCBServerIOStatus
 *                         enumerators that correspond to the contents
 *                         of rfds and wfds.
 */
static unsigned ccb_server_select_to_io(CCBServerLink *sl, const fd_set *rfds,
					const fd_set *wfds)
{
  unsigned io;  /* The bitmask set of CCBServerIOStatus enumerators */
/*
 * Check the arguments.
 */
  if(!sl || !rfds || !wfds) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(51),
	"ccb_server_select_to_io: NULL argument(s).\n");    
    return 0;  /* No events */
  };
/*
 * Translate the contents of rfds and wfds to the corresponding bitmask
 * union of CCBServerIOStatus enumerators.
 */
  io = 0;
  if(sl->cntrl_sock >= 0) {
    if(FD_ISSET(sl->cntrl_sock, rfds))
      io |= CCB_RD_CNTRL_SOCK;
    if(FD_ISSET(sl->cntrl_sock, wfds))
      io |= CCB_WR_CNTRL_SOCK;
  };
  if(sl->telem_sock >= 0) {
    if(FD_ISSET(sl->telem_sock, rfds))
      io |= CCB_RD_TELEM_SOCK;
    if(FD_ISSET(sl->telem_sock, wfds))
      io |= CCB_WR_TELEM_SOCK;
  };
  if(FD_ISSET(sl->cntrl_port, rfds))
    io |= CCB_RD_CNTRL_PORT;
  if(FD_ISSET(sl->telem_port, rfds))
    io |= CCB_RD_TELEM_PORT;
/*
 * Return the set of I/O events.
 */
  return io;
}

/*.......................................................................
 * Select the set of telemetry streams that are to be sent to the manager.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 *  streams    unsigned    The set of desired streams, expressed as
 *                         a bitmask union of CCBTelemetryStream
 *                         enumerators.
 * Output
 *  return          int    0 - OK.
 *                         1 - Error.
 */
int ccb_select_streams(CCBServerLink *sl, unsigned streams)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(52),
		"ccb_select_streams: NULL argument(s).\n");
    return 1;
  };
/*
 * Forward the request to the telemetry server.
 */
  return ccb_ts_select_streams(sl->ts, streams);
}

/*.......................................................................
 * Return the general status of the CCB for use in replies to
 * status-request commands.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 * Output:
 *  return     unsigned    The bitmask union of CCBGeneralStatus enumerators
 *                         which indicate the general status of the CCB.
 */
unsigned ccb_general_status(CCBServerLink *sl)
{
  unsigned status = 0;  /* The return value */
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(53),
		"ccb_general_status: NULL argument(s).\n");
    status = CCB_SOFT_FAULT;
  } else {
/*
 * Is the telemetry link down?
 */
    if(ccb_ts_status(sl->ts) == CCB_TELEM_SERVER_DOWN)
      status |= CCB_LINK_DOWN;
/*
 * Were we forced to close the integrated-data queue while the current
 * contents drain?
 */
    if(ccb_ts_draining(sl->ts))
      status |= CCB_BUFFER_FULL;
/*
 * Have any software or hardware faults been reported?
 */
    status |= sl->faults & (CCB_HARD_FAULT | CCB_SOFT_FAULT);
  };
/*
 * Return the set of status enumerators collected above.
 */
  return status;
}

/*.......................................................................
 * Record the detection of a software and/or hardware fault.
 *
 * Input:
 *  sl    CCBServerLink *  The networking interface of the CCB server.
 *  faults     unsigned    The bitwise union of zero or more of,
 *                          CCB_HARD_FAULT - A hardware fault was
 *                                           detected.
 *                          CCB_SOFT_FAULT - A software fault was
 *                                           detected.
 *                         This bitmask overrides any previously
 *                         specified list of faults. This allows
 *                         previously reported faults to be cleared.
 * Output:
 *  return          int    0 - OK.
 *                         1 - Error.
 */
int ccb_record_server_faults(CCBServerLink *sl, unsigned faults)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(54),
		"ccb_record_server_faults: NULL argument(s).\n");
    return 1;
  };
/*
 * Copy valid parts of the fault-mask.
 */
  sl->faults = faults & (CCB_HARD_FAULT | CCB_SOFT_FAULT);
  return 0;
}

/*.......................................................................
 * Set the period with which the server sub-samples the monitoring
 * packets.
 *
 * Input:
 *  sl        CCBServerLink *  The networking interface of the CCB server.
 *  period   unsigned short    The monitoring period, expressed as a
 *                             multiple of the integration period.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_monitor_period(CCBServerLink *sl, unsigned short period)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(55),
		"ccb_monitor_period: NULL argument(s).\n");
    return 1;
  };
/*
 * Forward the request to the telemetry server.
 */
  return ccb_ts_monitor_period(sl->ts, period);
}

/*.......................................................................
 * Set the interval at which historical log messages are forgotten.
 *
 * Input:
 *  sl        CCBServerLink *  The networking interface of the CCB server.
 *  period    unsigned long    The lifetime of the record of historical
 *                             log messages (seconds).
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
int ccb_log_msg_lifetime(CCBServerLink *sl, unsigned long period)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(56),
		"ccb_log_msg_lifetime: NULL argument(s).\n");
    return 1;
  };
/*
 * Forward the request to the telemetry server.
 */
  return ccb_ts_log_msg_lifetime(sl->ts, period);
}

/*-----------------------------------------------------------------------*
 *                     Driver control functions                          *
 *-----------------------------------------------------------------------*/

/*.......................................................................
 * Set the configuration parameters of the next scan.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 *  cnf  const CCBConfig *   The new CCB configuration to be installed.
 * Output:
 *  return           int     0 - OK.
 *                           1 - Error.
 */
int ccb_driver_conf_scan(CCBServerLink *sl, const CCBConfig *cnf)
{
  CCBDriverCmd cmd;   /* The command to be sent to the CCB device driver */
/*
 * Check the arguments.
 */
  if(!sl || !cnf) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(58),
		       "ccb_driver_conf_scan: NULL argument(s)\n");
    return 1;
  };
/*
 * We need a device driver to forward this request to.
 */
  if(ccb_driver_not_loaded(sl))
    return 1;
/*
 * Construct a command to be sent to the device driver.
 */
  cmd.type = CCB_DRV_CONF_SCAN;
  if(ccb_get_phase_switch_cnf(cnf, &cmd.pars.conf.phase) ||
     ccb_get_cal_diode_cnf(cnf, &cmd.pars.conf.cal) ||
     ccb_get_timing_cnf(cnf, &cmd.pars.conf.timing) ||
     ccb_get_sampler_cnf(cnf, &cmd.pars.conf.sampler))
    return 1;
/*
 * Send the command to the device driver.
 */
  return sl->driver->tell_driver(sl, sl->driver->data, &cmd);
}

/*.......................................................................
 * Start a new scan on a given 1PPS-tick.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 *  scan   unsigned long     The numeric scan identifier to record in
 *                           integrated data messages associated with
 *                           the new scan.
 *  mjd    unsigned long     The day on which to start the scan,
 *                           expressed as a Modified Julian Day number.
 *  secs   unsigned long     The time-of-day at which to start the scan,
 *                           expressed as the number of seconds from the
 *                           start of the specified day.
 * Output:
 *  return           int     0 - OK.
 *                           1 - Error.
 */
int ccb_driver_stage_scan(CCBServerLink *sl, unsigned long scan,
			  unsigned long mjd, unsigned long secs)
{
  CCBDriverCmd cmd;   /* The command to be sent to the CCB device driver */
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(59),
		       "ccb_driver_stage_scan: NULL argument(s)\n");
    return 1;
  };
/*
 * We need a device driver to forward this request to.
 */
  if(ccb_driver_not_loaded(sl))
    return 1;
/*
 * Construct a command to be sent to the device driver.
 */
  cmd.type = CCB_DRV_STAGE_SCAN;
  cmd.pars.stage.scan = scan;
  cmd.pars.stage.mjd = mjd;
  cmd.pars.stage.secs = secs;
/*
 * Send the command to the device driver.
 */
  return sl->driver->tell_driver(sl, sl->driver->data, &cmd);
}

/*.......................................................................
 * Start a new intra-scan as soon as possible.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 *  scan   unsigned long     The numeric scan identifier to record in
 *                           integrated data messages associated with
 *                           the new scan.
 * Output:
 *  return           int     0 - OK.
 *                           1 - Error.
 */
int ccb_driver_intra_scan(CCBServerLink *sl, unsigned long scan)
{
  CCBDriverCmd cmd;   /* The command to be sent to the CCB device driver */
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(60),
		       "ccb_driver_intra_scan: NULL argument(s)\n");
    return 1;
  };
/*
 * We need a device driver to forward this request to.
 */
  if(ccb_driver_not_loaded(sl))
    return 1;
/*
 * Construct a command to be sent to the device driver.
 */
  cmd.type = CCB_DRV_INTRA_SCAN;
  cmd.pars.intra.scan = scan;
/*
 * Send the command to the device driver.
 */
  return sl->driver->tell_driver(sl, sl->driver->data, &cmd);
}

/*.......................................................................
 * Start a new dump-mode intra-scan as soon as possible.
 *
 * Input:
 *  sl       CCBServerLink *  The networking interface of the CCB server.
 *  scan     unsigned long    The numeric ID to give the new intra-scan.
 *  adc     unsigned short    The ID of the ADC whose samples are to be
 *                            collected (0-15).
 *  samples  unsigned long    The max number of samples to collect per
 *                            integration.
 *  frames   unsigned long    The number of per-integration data-frames
 *                            to deliver.
 * Output:
 *  return             int    0 - OK.
 *                            1 - Error.
 */
int ccb_driver_dump_scan(CCBServerLink *sl, unsigned long scan,
			 unsigned short adc, unsigned long samples,
			 unsigned long frames)
{
  CCBDriverCmd cmd;   /* The command to be sent to the CCB device driver */
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(61),
		       "ccb_driver_dump_scan: NULL argument(s)\n");
    return 1;
  };
/*
 * We need a device driver to forward this request to.
 */
  if(ccb_driver_not_loaded(sl))
    return 1;
/*
 * Construct a command to be sent to the device driver.
 */
  cmd.type = CCB_DRV_DUMP_SCAN;
  cmd.pars.dump.scan = scan;
  cmd.pars.dump.adc = adc;
  cmd.pars.dump.samples = samples;
  cmd.pars.dump.frames = frames;
/*
 * Send the command to the device driver.
 */
  return sl->driver->tell_driver(sl, sl->driver->data, &cmd);
}

/*.......................................................................
 * After first unloading any existing CCB device-driver, load the CCB
 * device driver anew, then reset both the server and the driver to
 * power-on defaults. The CCB is left in standby mode, performing an
 * intra-scan of ID 0.
 *
 * Input:
 *  sl        CCBServerLink *   The networking interface of the CCB server.
 *  type      CCBDriverType     The type of driver to load.
 * Output:
 *  return              int     0 - OK.
 *                              1 - Error.
 */
int ccb_load_driver(CCBServerLink *sl, CCBDriverType type)
{
  CCBServerDriver *driver = NULL;  /* The driver to load */
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(62),
		       "ccb_load_driver: NULL argument(s)\n");
    return 1;
  };
/*
 * Ensure that the driver isn't already loaded.
 */
  if(ccb_unload_driver(sl))
    return 1;
/*
 * Clear any CCB fault flags.
 */
  ccb_record_server_faults(sl, 0);
/*
 * Reset the control and telemetry servers to their power-on defaults.
 */
  if(ccb_cs_reset(sl->cs) || ccb_ts_reset(sl->ts))
    return 1;
/*
 * Lookup the requested driver.
 */
  switch(type) {
  case CCB_NORMAL_DRIVER:
    driver = sl->normal;
    break;
  case CCB_VIRTUAL_DRIVER:
    driver = sl->virtual;
    break;
  default:
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(63),
		       "ccb_load_driver: Unknown driver type requested.\n");
    return 1;
  };
/*
 * Is the requested driver not-implemented yet?
 */
  if(!driver) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(64),
       "The requested device driver doesn't appear to be implemented yet.\n");
    return 1;
  };
/*
 * Forward the request to the application.
 */
  if(driver->load_driver(sl, driver->data)) {
    driver->unload_driver(sl, driver->data);
    return 1;
  };
/*
 * Install the interface object of the successfully loaded driver.
 */
  sl->driver = driver;
/*
 * Start with the telemetry stream only sending log-messages.
 */
  if(ccb_ts_select_streams(sl->ts, CCB_LOG_STREAM))
    return 1;
/*
 * Install the initial configuration of the CCB and start a dummy intra-scan.
 */
  if(ccb_cs_intra_scan(sl->cs, 0))
    return 1;
/*
 * Initiate a dummy intra-scan.
 */
  return 0;
}

/*.......................................................................
 * Unload the CCB device driver. Note that this function may be called
 * when no driver is currently loaded, and that this shouldn't be
 * interpreted as an error.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 * Output:
 *  return           int     0 - OK.
 *                           1 - Error.
 */
int ccb_unload_driver(CCBServerLink *sl)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(65),
		       "ccb_unload_driver: NULL argument(s)\n");
    return 1;
  };
/*
 * Is there a driver to be unloaded?
 */
  if(sl->driver) {
/*
 * Forward the request to the driver interface.
 */
    sl->driver->unload_driver(sl, sl->driver->data);
/*
 * Uninstall the driver interface object.
 */
    sl->driver = NULL;
  };
  return 0;
}

/*.......................................................................
 * Reset the current device-driver (if any), by unloading and reloading
 * it.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 * Output:
 *  return           int     0 - OK.
 *                           1 - Error.
 */
int ccb_reset_driver(CCBServerLink *sl)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(66),
		       "ccb_reset_driver: NULL argument(s)\n");
    return 1;
  };
/*
 * If no driver is currently loaded, do nothing.
 */
  if(!sl->driver)
    return 0;
/*
 * Identify the driver that is currently loaded, and reload it.
 */
  if(sl->driver == sl->normal) {
    return ccb_load_driver(sl, CCB_NORMAL_DRIVER);
  } else if(sl->driver == sl->virtual) {
    return ccb_load_driver(sl, CCB_VIRTUAL_DRIVER);
  } else {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(67),
		       "ccb_reset_driver: Unknown driver.\n");
    return 1;
  };    
  return 0;
}

/*.......................................................................
 * Reboot the CCB real-time computer, or simply arrange that it be
 * rebooted ASAP.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 * Output:
 *  return           int     0 - OK.
 *                           1 - Error.
 */
int ccb_reboot_rtc(CCBServerLink *sl)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(68),
		       "ccb_reboot_rtc: NULL argument(s)\n");
    return 1;
  };
/*
 * We need a device driver to forward this request to.
 */
  if(ccb_driver_not_loaded(sl))
    return 1;
/*
 * Forward the request to the application.
 */
  return sl->driver->reboot_rtc(sl, sl->driver->data);
}

/*.......................................................................
 * Shutdown the CCB real-time computer, or simply arrange that it be
 * shutdown ASAP.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 * Output:
 *  return           int     0 - OK.
 *                           1 - Error.
 */
int ccb_shutdown_rtc(CCBServerLink *sl)
{
/*
 * Check the arguments.
 */
  if(!sl) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(69),
		       "ccb_shutdown_rtc: NULL argument(s)\n");
    return 1;
  };
/*
 * We need a device driver to forward this request to.
 */
  if(ccb_driver_not_loaded(sl))
    return 1;
/*
 * Forward the request to the application.
 */
  return sl->driver->shutdown_rtc(sl, sl->driver->data);
}

/*.......................................................................
 * Check for, and report an error if no device driver is currently loaded.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 * Output:
 *  return           int     0 - The driver is loaded.
 *                           1 - The driver is not-loaded.
 */
static int ccb_driver_not_loaded(CCBServerLink *sl)
{
  if(!sl || !sl->driver) {
    ccb_log_server_msg(sl, CCB_ERROR, CCB_LOGID(70),
		       "No CCB device-driver is currently loaded.\n");
    return 1;
  };
  return 0;
}

/*.......................................................................
 * This is the I/O-driven event loop that is used to watch for and react
 * to client and driver communications.
 *
 * Input:
 *  sl     CCBServerLink *   The networking interface of the CCB server.
 * Output:
 *  return           int    0 - OK.
 *                          1 - Error.
 */
int ccb_server_event_loop(CCBServerLink *sl)
{
/*
 * The event loop runs until an error occurs.
 */
  while(1) {
    fd_set rfds;    /* The set of file-descriptors to watch for readability */
    fd_set wfds;    /* The set of file-descriptors to watch for writability */
    int maxfd=0;    /* The maximum of the file-descriptors to watch */
    int nready;     /* The number of file descriptors ready for I/O */
    struct timeval timeout;  /* The select() inactivity-timeout */
/*
 * Construct the sets of file descriptors to be watched from scratch.
 */
    FD_ZERO(&rfds);
    FD_ZERO(&wfds);
/*
 * Set the timeout to a bogus value, so that we can check whether
 * the driver installs a timeout or not.
 */
    timeout.tv_sec = -1;
    timeout.tv_usec = 0;
/*
 * Add the file-descriptors that the driver wishes to have watched.
 */
    if(sl->driver && sl->driver->select_events(sl, sl->driver->data, &rfds,
					       &wfds, &maxfd, &timeout))
      return 1;
/*
 * Add the file-descriptors that the server needs to have watched.
 */
    if(ccb_server_select_args(sl, &rfds, &wfds, &maxfd))
      return 1;
/*
 * Wait for I/O events and inactivity timeouts.
 */
    nready = select(maxfd + 1, &rfds, &wfds, NULL,
		    (timeout.tv_sec == -1 ? NULL : &timeout));
/*
 * Error?
 */
    if(nready < 0) {
      if(errno != EINTR) {
	fprintf(stderr, "select() error (%s).\n", strerror(errno));
	return 1;
      };
/*
 * Timeout?
 */
    } else if(nready == 0) {
/*
 * Tell the driver about the timeout.
 */
      if(sl->driver && sl->driver->check_events(sl, sl->driver->data,
						NULL, NULL))
	return 1;
/*
 * At least one file-descriptor is ready for I/O.
 */
    } else {
/*
 * Perform the types of server I/O indicated by select().
 */
      if(ccb_server_communicate(sl, &rfds, &wfds))
	return 1;
/*
 * Now allow the driver to respond to activity on its file-descriptors.
 */
      if(sl->driver && sl->driver->check_events(sl, sl->driver->data,
						&rfds, &wfds))
	return 1;
    };
  };
/*
 * This statement shouldn't be reached.
 */
  return 0;
}

