#ifndef ccbserverlink_h
#define ccbserverlink_h

/*-----------------------------------------------------------------------*
 * This is the public header file of the client portion of the CCB       *
 * communications library.                                               *
 *-----------------------------------------------------------------------*/

/*
 * Get datatypes that are common to both the server and client sides of
 * the CCB communications library.
 */
#include "ccbcommon.h"

#ifndef __KERNEL__
#include <sys/types.h>   /* fd_set */
#include <stdarg.h>
#endif

/*-----------------------------------------------------------------------
 * The following declarations are used not only by the server, but also
 * by the kernel device driver.
 *-----------------------------------------------------------------------*/

/*
 * Enumerate all device-driver commands sent by the server.
 */
typedef enum {
  CCB_DRV_CONF_SCAN,   /* Install a new configuration */
  CCB_DRV_STAGE_SCAN,  /* Start a new scan at the specified time */
  CCB_DRV_INTRA_SCAN,  /* Start an intra-scan ASAP */
  CCB_DRV_DUMP_SCAN    /* Start a dump-mode intra-scan ASAP */
} CCBDriverCmdID;

/*
 * This is the body of a CCB_DRV_CONF_SCAN command.
 */
typedef struct {
  CCBPhaseSwitchCnf phase;  /* The phase-switch configuration */
  CCBCalDiodeCnf cal;       /* The calibration diode configuration */
  CCBTimingCnf timing;      /* The hardware-timing configuration */
  CCBSamplerCnf sampler;    /* The sampler configuration */
} CCBDrvConfScan;

/*
 * This is the body of the CCB_DRV_STAGE_SCAN command.
 */
typedef struct {
  unsigned long scan; /* The numeric scan identifier */
  unsigned long mjd;  /* The date on which to start, expressed as a Modified */
                      /*  Julian day number. */
  unsigned long secs; /* The time of day at which to start the scan, */
                      /*  expressed in seconds since 0H UTC on the */
                      /*  day specified in 'mjd'. */
} CCBDrvStageScan;

/*
 * This is the body of the CCB_DRV_INTRA_SCAN command.
 */
typedef struct {
  unsigned long scan; /* The numeric scan identifier */
} CCBDrvIntraScan;

/*
 * This is the body of the CCB_DRV_DUMP_SCAN command.
 */
typedef struct {
  unsigned long scan;     /* The numeric scan identifier */
  unsigned short adc;     /* The ADC to dump */
  unsigned long samples;  /* The max number of samples to */
                          /*  collect per integration. */
  unsigned long frames;   /* The number of per-integration */
                          /*  data-frames to deliver. */
} CCBDrvDumpScan;

/*
 * Define the container which is used to send commands to the
 * device driver.
 */
typedef struct {
  CCBDriverCmdID type;         /* The type of command being sent */
  union {
    CCBDrvConfScan conf;       /* type==CCB_DRV_CONF_SCAN */
    CCBDrvStageScan stage;     /* type==CCB_DRV_STAGE_SCAN */
    CCBDrvIntraScan intra;     /* type==CCB_DRV_INTRA_SCAN */
    CCBDrvDumpScan dump;       /* type==CCB_DRV_DUMP_SCAN */
  } pars;
} CCBDriverCmd;

/*-----------------------------------------------------------------------
 * The following declarations can not be used by the kernel device-driver.
 *-----------------------------------------------------------------------*/

#ifndef __KERNEL__
/*
 * The CCBServerLink datatype holds the dynamically-allocated resources
 * that CCB server communications with a CCB manager. The contents of
 * the structure are private to the library, so this public header
 * file only includes an opaque declaration.
 */
typedef struct CCBServerLink CCBServerLink;

/*
 * The CCBServerDriver datatype encapsulates the method-functions
 * that provide the server's interface to a given CCB device driver.
 * The contents of the structure are private to the library, so this
 * public header file only includes an opaque declaration.
 */
typedef struct CCBServerDriver CCBServerDriver;

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to load the CCB device driver.  Note that this function
 * is only called after a preceding call to ccb_unload_driver().
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_LOAD_FN(fn) int (fn)(CCBServerLink *sl, void *data)

typedef CCB_DRIVER_LOAD_FN(CCBDriverLoadFn);

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to unload the CCB device driver.  Beware that this
 * function will sometimes be called when no device driver is currently
 * loaded. This should not be treated as an error.
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_UNLOAD_FN(fn) int (fn)(CCBServerLink *sl, void *data)

typedef CCB_DRIVER_UNLOAD_FN(CCBDriverUnloadFn);


/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to send a command to the CCB device driver.
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 *  cmd  const CCBDriverCmd *  The command to be sent to the driver.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_TELL_FN(fn) int (fn)(CCBServerLink *sl, void *data, \
					const CCBDriverCmd *cmd)

typedef CCB_DRIVER_TELL_FN(CCBDriverTellFn);

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to reboot the CCB real-time computer. The function need
 * not wait for the reboot process to actually start.
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_REBOOT_RTC_FN(fn) int (fn)(CCBServerLink *sl, void *data)

typedef CCB_REBOOT_RTC_FN(CCBRebootRTCFn);

/*.......................................................................
 * A method function of this type is called whenever a CCBServerLink
 * object needs to shutdown the CCB real-time computer. The function need
 * not wait for the shutdown process to actually start.
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_SHUTDOWN_RTC_FN(fn) int (fn)(CCBServerLink *sl, void *data)

typedef CCB_SHUTDOWN_RTC_FN(CCBShutdownRTCFn);

/*.......................................................................
 * This driver method-function is called just before the server
 * calls select() to wait for I/O, in order that the driver can register
 * the file-descriptors that it wants to have watched for I/O, and
 * optionally specify a select() inactivity timeout.
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 *  rfds             fd_set *  The callback should use the FD_SET() macro to
 *                             add any file-descriptors that it wants to have
 *                             watched for readability, to this set.
 *  wfds             fd_set *  The callback should use the FD_SET() macro to
 *                             add any file-descriptors that it wants to have
 *                             watched for writability, to this set.
 *  maxfd               int *  If *maxfd is < the maximum of the file
 *                             descriptors added by this function to
 *                             rfds or wfds, then it should updated to
 *                             have the latter value.
 *  timeout  struct timeval *  If the driver needs to set an inactivity
 *                             timeout, it should update the contents of
 *                             this structure. Otherwise it should leave
 *                             it unchanged, and the server will see this
 *                             and not install any inactivity-timeout.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_SELECT_EVENTS_FN(fn) int (fn)(CCBServerLink *sl, \
   void *data, fd_set *rfds, fd_set *wfds, int *maxfd, struct timeval *timeout)

typedef CCB_DRIVER_SELECT_EVENTS_FN(CCBDriverSelectEventsFn);

/*.......................................................................
 * This driver method-function is called after the server's call to
 * select() returns, either due to file-descriptor activity or an
 * inactivity timeout. The latter case is signalled by rfds and wfds
 * being NULL.
 *
 * Input:
 *  sl        CCBServerLink *  The parent CCB server.
 *  data               void *  The application-specific callback data
 *                             that was passed in the synonymous argument
 *                             of new_CCBServerLink().
 *  rfds             fd_set *  This is NULL if a timeout occurred. Otherwise
 *                             it indicates the set of file-descriptors
 *                             that are now readable.
 *  wfds             fd_set *  This is NULL if a timeout occurred. Otherwise
 *                             it indicates the set of file-descriptors
 *                             that are now writable.
 * Output:
 *  return              int    0 - OK.
 *                             1 - Error.
 */
#define CCB_DRIVER_CHECK_EVENTS_FN(fn) int (fn)(CCBServerLink *sl, \
			void *data, fd_set *rfds, fd_set *wfds)
typedef CCB_DRIVER_CHECK_EVENTS_FN(CCBDriverCheckEventsFn);

/*
 * Create an object which encapsulates the method-functions that are used
 * to talk to a given CCB device-driver.
 */
CCBServerDriver *new_CCBServerDriver(void *data,
				     CCBDriverLoadFn *load_driver,
				     CCBDriverUnloadFn *unload_driver,
				     CCBDriverTellFn *tell_driver,
				     CCBDriverSelectEventsFn *select_events,
				     CCBDriverCheckEventsFn *check_events,
				     CCBRebootRTCFn *reboot_rtc,
				     CCBShutdownRTCFn *shutdown_rtc);
/*
 * Delete an object that was previously returned by new_CCBServerDriver().
 */
CCBServerDriver *del_CCBServerDriver(CCBServerDriver *drv);

/*
 * Create the resources needed to communicate with a remote CCB manager
 * process.
 */
CCBServerLink *new_CCBServerLink(CCBServerDriver *normal,
				 CCBServerDriver *virtual);

/*
 * Shutdown the communications of any connected manager, then return
 * the resources of the specified CCBServerLink object to the system.
 */
CCBServerLink *del_CCBServerLink(CCBServerLink *sl);

/*
 * Invoke the CCB server's event-loop.
 */
int ccb_server_event_loop(CCBServerLink *sl);

/*
 * Queue an integ-data message to be sent to the current manager.
 */
int ccb_queue_integ_msg(CCBServerLink *sl, const CCBTimeStamp *t,
			unsigned long scan, unsigned long number,
			unsigned short flags, const unsigned long *values,
			unsigned nvalues);
/*
 * Queue a monitor-data message to be sent to the current manager.
 */
int ccb_queue_monitor_msg(CCBServerLink *sl, const CCBTimeStamp *t,
			  unsigned long scan, unsigned long number,
			  const unsigned long *values, unsigned nvalues);

/* The following function is used for reporting all CCB server log messages */

int ccb_log_server_msg(CCBServerLink *sl, CCBLogLevel level, unsigned long id,
		       const char *fmt, ...)
#ifdef __GNUC__                          /* Arrange for gcc to check the */
__attribute__ ((format (printf, 4, 5))); /* printf-style format and arguments */
#else
 ;
#endif

/*
 * Record the existence of software and/or hardware faults.
 */
int ccb_record_server_faults(CCBServerLink *sl, unsigned faults);

#endif

#endif
